from __future__ import annotations

from collections.abc import Callable, Sequence
from copy import deepcopy
from datetime import datetime
from typing import Any, Literal
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from bluesky.run_engine import RunEngine
from bluesky.utils import MsgGenerator
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.synchrotron import SynchrotronMode
from ophyd_async.core import set_mock_value

from mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan import (
    grid_detect_then_xray_centre,
)
from mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping import (
    populate_data_collection_group,
    populate_remaining_data_collection_info,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_mapping import (
    construct_comment_for_gridscan,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    DataCollectionInfo,
    Orientation,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.parameters.components import IspybExperimentType
from mx_bluesky.common.parameters.rotation import (
    RotationScan,
)
from mx_bluesky.hyperion.experiment_plans.hyperion_flyscan_xray_centre_plan import (
    construct_hyperion_specific_features,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
    rotation_scan_internal,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    RotationISPyBCallback,
)
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionGridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import (
    GridCommonWithHyperionDetectorParams,
    GridScanWithEdgeDetect,
    HyperionSpecifiedThreeDGridScan,
)

from ....conftest import SimConstants, replace_all_tmp_paths
from ...conftest import (
    DATA_COLLECTION_COLUMN_MAP,
    compare_actual_and_expected,
    compare_comment,
)
from .conftest import raw_params_from_file

EXPECTED_DATACOLLECTION_FOR_ROTATION = {
    "wavelength": 0.71,
    "beamSizeAtSampleX": 0.02,
    "beamSizeAtSampleY": 0.02,
    "exposureTime": 0.023,
    "undulatorGap1": 1.11,
    "synchrotronMode": SynchrotronMode.USER.value,
    "slitGapHorizontal": 0.123,
    "slitGapVertical": 0.234,
}

GRID_INFO_COLUMN_MAP = {
    s.lower(): s
    for s in [
        "gridInfoId",
        "dataCollectionGroupId",
        "xOffset",
        "yOffset",
        "dx_mm",
        "dy_mm",
        "steps_x",
        "steps_y",
        "meshAngle",
        "micronsPerPixelX",
        "micronsPerPixelY",
        "snapshot_offsetXPixel",
        "snapshot_offsetYPixel",
        "recordTimeStamp",
        "orientation",
        "workflowMeshId",
        "snaked",
        "dataCollectionId",
        "patchesX",
        "patchesY",
        "micronsPerPixelX",
        "micronsPerPixelY",
    ]
}


@pytest.fixture
def dummy_data_collection_group_info(dummy_params):
    return populate_data_collection_group(
        dummy_params,
    )


@pytest.fixture
def dummy_scan_data_info_for_begin_xy(dummy_params):
    info = DataCollectionInfo(
        data_collection_number=dummy_params.detector_params.run_number,
    )
    info = populate_remaining_data_collection_info(
        "MX-Bluesky: Xray centring 1 -", None, info, dummy_params
    )
    return ScanDataInfo(
        data_collection_info=info,
    )


@pytest.fixture
def dummy_scan_data_info_for_begin_xz(dummy_params):
    run_number = dummy_params.detector_params.run_number + 1
    info1 = DataCollectionInfo(
        data_collection_number=run_number,
    )
    info = info1
    info = populate_remaining_data_collection_info(
        "MX-Bluesky: Xray centring 2 -", None, info, dummy_params
    )
    return ScanDataInfo(
        data_collection_info=info,
    )


@pytest.fixture
def storage_directory(tmp_path) -> str:
    return str(tmp_path)


@pytest.fixture
def grid_detect_then_xray_centre_parameters(tmp_path):
    json_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/ispyb_gridscan_system_test_parameters.json",
        tmp_path,
    )
    json_dict["sample_id"] = SimConstants.ST_SAMPLE_ID
    json_dict["visit"] = SimConstants.ST_VISIT
    return GridScanWithEdgeDetect(**json_dict)


def scan_xy_data_info_for_update(
    data_collection_group_id,
    dummy_params: HyperionSpecifiedThreeDGridScan,
    scan_data_info_for_begin,
):
    scan_data_info_for_update = deepcopy(scan_data_info_for_begin)
    scan_data_info_for_update.data_collection_info.parent_id = data_collection_group_id
    assert dummy_params is not None
    scan_data_info_for_update.data_collection_grid_info = DataCollectionGridInfo(
        dx_in_mm=dummy_params.x_step_size_um,
        dy_in_mm=dummy_params.y_step_size_um,
        steps_x=dummy_params.x_steps,
        steps_y=dummy_params.y_steps,
        microns_per_pixel_x=1.25,
        microns_per_pixel_y=1.25,
        # cast coordinates from numpy int64 to avoid mysql type conversion issues
        snapshot_offset_x_pixel=100,
        snapshot_offset_y_pixel=100,
        orientation=Orientation.HORIZONTAL,
        snaked=True,
    )
    scan_data_info_for_update.data_collection_info.comments = (
        construct_comment_for_gridscan(
            scan_data_info_for_update.data_collection_grid_info,
        )
    )
    return scan_data_info_for_update


def scan_data_infos_for_update_3d(
    ispyb_ids,
    scan_xy_data_info_for_update,
    dummy_params: HyperionSpecifiedThreeDGridScan,
):
    run_number = dummy_params.detector_params.run_number + 1
    info = DataCollectionInfo(
        data_collection_number=run_number,
    )
    xz_data_collection_info = info

    assert dummy_params is not None
    data_collection_grid_info = DataCollectionGridInfo(
        dx_in_mm=dummy_params.x_step_size_um,
        dy_in_mm=dummy_params.z_step_size_um,
        steps_x=dummy_params.x_steps,
        steps_y=dummy_params.z_steps,
        microns_per_pixel_x=1.25,
        microns_per_pixel_y=1.25,
        # cast coordinates from numpy int64 to avoid mysql type conversion issues
        snapshot_offset_x_pixel=100,
        snapshot_offset_y_pixel=50,
        orientation=Orientation.HORIZONTAL,
        snaked=True,
    )
    xz_data_collection_info = populate_remaining_data_collection_info(
        construct_comment_for_gridscan(data_collection_grid_info),
        ispyb_ids.data_collection_group_id,
        xz_data_collection_info,
        dummy_params,
    )
    xz_data_collection_info.parent_id = ispyb_ids.data_collection_group_id

    scan_xz_data_info_for_update = ScanDataInfo(
        data_collection_id=ispyb_ids.data_collection_ids[1],
        data_collection_info=xz_data_collection_info,
        data_collection_grid_info=(data_collection_grid_info),
    )
    return [scan_xy_data_info_for_update, scan_xz_data_info_for_update]


@pytest.mark.system_test
def test_ispyb_deposition_comment_correct_on_failure(
    dummy_ispyb: StoreInIspyb,
    fetch_comment: Callable[..., Any],
    dummy_data_collection_group_info,
    dummy_scan_data_info_for_begin_xy,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_data_collection_group_info, [dummy_scan_data_info_for_begin_xy]
    )
    dummy_ispyb.end_deposition(ispyb_ids, "fail", "could not connect to devices")
    assert (
        fetch_comment(ispyb_ids.data_collection_ids[0])  # type: ignore
        == "MX-Bluesky: Xray centring 1 - DataCollection Unsuccessful reason: could not connect to devices"
    )


@patch("mx_bluesky.common.external_interaction.ispyb.ispyb_utils.datetime")
@pytest.mark.system_test
def test_ispyb_deposition_comment_handles_long_comment_and_commits_end_status(
    mock_datetime: MagicMock,
    dummy_params,
    dummy_ispyb: StoreInIspyb,
    fetch_datacollection_attribute: Callable[..., Any],
    dummy_data_collection_group_info,
    dummy_scan_data_info_for_begin_xy,
):
    timestamp = datetime.fromisoformat("2024-08-11T15:59:23")
    mock_datetime.datetime = MagicMock(**{"now.return_value": timestamp})  # type: ignore
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_data_collection_group_info, [dummy_scan_data_info_for_begin_xy]
    )
    dummy_ispyb.end_deposition(
        ispyb_ids, "fail", f"Failed with very big object repr {dummy_params}"
    )

    expected_values = {"endTime": timestamp, "runStatus": "DataCollection Unsuccessful"}
    compare_actual_and_expected(
        ispyb_ids.data_collection_ids[0],
        expected_values,
        fetch_datacollection_attribute,
    )


@pytest.mark.system_test
def test_ispyb_deposition_comment_correct_for_3d_on_failure(
    dummy_ispyb: StoreInIspyb,
    fetch_comment: Callable[..., Any],
    dummy_params,
    dummy_data_collection_group_info,
    dummy_scan_data_info_for_begin_xy,
    dummy_scan_data_info_for_begin_xz,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_data_collection_group_info,
        [dummy_scan_data_info_for_begin_xy, dummy_scan_data_info_for_begin_xz],
    )
    scan_data_infos = generate_scan_data_infos(
        dummy_params,
        dummy_scan_data_info_for_begin_xy,
        IspybExperimentType.GRIDSCAN_3D,
        ispyb_ids,
    )
    ispyb_ids = dummy_ispyb.update_deposition(ispyb_ids, scan_data_infos)
    dcid1 = ispyb_ids.data_collection_ids[0]  # type: ignore
    dcid2 = ispyb_ids.data_collection_ids[1]  # type: ignore
    dummy_ispyb.end_deposition(ispyb_ids, "fail", "could not connect to devices")
    assert (
        fetch_comment(dcid1)
        == "MX-Bluesky: Xray centring 1 - Diffraction grid scan of 40 by 20 images in 100.0 um by 100.0 um steps. Top "
        "left (px): [100,100], bottom right (px): [3300,1700]. DataCollection Unsuccessful reason: could not connect to devices"
    )
    assert (
        fetch_comment(dcid2)
        == "MX-Bluesky: Xray centring 2 - Diffraction grid scan of 40 by 10 images in 100.0 um by 100.0 um steps. Top "
        "left (px): [100,50], bottom right (px): [3300,850]. DataCollection Unsuccessful reason: could not connect to devices"
    )


@pytest.mark.system_test
@pytest.mark.parametrize(
    "experiment_type, exp_num_of_grids, success",
    [
        (IspybExperimentType.GRIDSCAN_2D, 1, False),
        (IspybExperimentType.GRIDSCAN_2D, 1, True),
        (IspybExperimentType.GRIDSCAN_3D, 2, False),
        (IspybExperimentType.GRIDSCAN_3D, 2, True),
    ],
)
def test_can_store_2d_ispyb_data_correctly_when_in_error(
    experiment_type,
    exp_num_of_grids: Literal[1, 2],
    success: bool,
    fetch_comment: Callable[..., Any],
    dummy_params,
    dummy_data_collection_group_info,
    dummy_scan_data_info_for_begin_xy,
    dummy_scan_data_info_for_begin_xz,
    ispyb_config_path: str,
):
    ispyb: StoreInIspyb = StoreInIspyb(ispyb_config_path)
    scan_data_infos = [dummy_scan_data_info_for_begin_xy]
    if experiment_type == IspybExperimentType.GRIDSCAN_3D:
        scan_data_infos += [dummy_scan_data_info_for_begin_xz]
    ispyb_ids: IspybIds = ispyb.begin_deposition(
        dummy_data_collection_group_info, scan_data_infos
    )
    scan_data_infos = generate_scan_data_infos(
        dummy_params, dummy_scan_data_info_for_begin_xy, experiment_type, ispyb_ids
    )

    ispyb_ids = ispyb.update_deposition(ispyb_ids, scan_data_infos)
    assert len(ispyb_ids.data_collection_ids) == exp_num_of_grids  # type: ignore
    assert len(ispyb_ids.grid_ids) == exp_num_of_grids  # type: ignore
    assert isinstance(ispyb_ids.data_collection_group_id, int)

    expected_comments = [
        (
            "MX-Bluesky: Xray centring 1 - Diffraction grid scan of 40 by 20 "
            "images in 100.0 um by 100.0 um steps. Top left (px): [100,100], bottom right (px): [3300,1700]."
        ),
        (
            "MX-Bluesky: Xray centring 2 - Diffraction grid scan of 40 by 10 "
            "images in 100.0 um by 100.0 um steps. Top left (px): [100,50], bottom right (px): [3300,850]."
        ),
    ]

    if success:
        ispyb.end_deposition(ispyb_ids, "success", "")
    else:
        ispyb.end_deposition(ispyb_ids, "fail", "In error")
        expected_comments = [
            e + " DataCollection Unsuccessful reason: In error"
            for e in expected_comments
        ]

    assert (
        not isinstance(ispyb_ids.data_collection_ids, int)
        and ispyb_ids.data_collection_ids is not None
    )
    for grid_no, dc_id in enumerate(ispyb_ids.data_collection_ids):
        assert fetch_comment(dc_id) == expected_comments[grid_no]


def test_ispyb_store_can_deal_with_data_collection_info_with_numpy_float64(
    dummy_params,
    dummy_data_collection_group_info,
    dummy_scan_data_info_for_begin_xy,
    dummy_scan_data_info_for_begin_xz,
    ispyb_config_path: str,
):
    ispyb: StoreInIspyb = StoreInIspyb(ispyb_config_path)
    experiment_type = IspybExperimentType.GRIDSCAN_3D
    scan_data_infos = [dummy_scan_data_info_for_begin_xy]
    if experiment_type == IspybExperimentType.GRIDSCAN_3D:
        scan_data_infos += [dummy_scan_data_info_for_begin_xz]
    ispyb_ids: IspybIds = ispyb.begin_deposition(
        dummy_data_collection_group_info, scan_data_infos
    )
    scan_data_infos = generate_scan_data_infos(
        dummy_params, dummy_scan_data_info_for_begin_xy, experiment_type, ispyb_ids
    )
    scan_data_infos[-1].data_collection_info.xbeam = np.float64(
        scan_data_infos[-1].data_collection_info.xbeam
    )
    scan_data_infos[-1].data_collection_info.ybeam = np.float64(
        scan_data_infos[-1].data_collection_info.ybeam
    )
    ispyb_ids = ispyb.update_deposition(ispyb_ids, scan_data_infos)


@pytest.mark.system_test
def test_ispyb_deposition_in_gridscan(
    run_engine: RunEngine,
    grid_detect_then_xray_centre_composite: HyperionGridDetectThenXRayCentreComposite,
    grid_detect_then_xray_centre_parameters: GridScanWithEdgeDetect,
    fetch_datacollection_attribute: Callable[..., Any],
    fetch_datacollection_grid_attribute: Callable[..., Any],
    fetch_datacollection_position_attribute: Callable[..., Any],
    storage_directory: str,
):
    set_mock_value(
        grid_detect_then_xray_centre_composite.s4_slit_gaps.xgap.user_readback, 0.1
    )
    set_mock_value(
        grid_detect_then_xray_centre_composite.s4_slit_gaps.ygap.user_readback, 0.1
    )
    ispyb_callback = GridscanISPyBCallback(GridCommonWithHyperionDetectorParams)
    run_engine.subscribe(ispyb_callback)
    run_engine(
        grid_detect_then_xray_centre(
            grid_detect_then_xray_centre_composite,
            grid_detect_then_xray_centre_parameters,
            HyperionSpecifiedThreeDGridScan,
            construct_hyperion_specific_features,
        )
    )

    ispyb_ids = ispyb_callback.ispyb_ids
    dc_expected_values = {
        "detectorid": 78,
        "axisstart": 0.0,
        "axisrange": 0,
        "axisend": 0,
        "slitgapvertical": 0.1,
        "slitgaphorizontal": 0.1,
        "beamsizeatsamplex": 0.02,
        "beamsizeatsampley": 0.02,
        "transmission": 49.118,
        "datacollectionnumber": 1,
        "detectordistance": 100.0,
        "exposuretime": 0.12,
        "imagedirectory": f"{storage_directory}/",
        "imageprefix": "file_name",
        "imagesuffix": "h5",
        "numberofpasses": 1,
        "overlap": 0,
        "omegastart": 0,
        "startimagenumber": 1,
        "wavelength": 0.976254,
        "xbeam": 150.0,
        "ybeam": 160.0,
        "xtalsnapshotfullpath1": f"{storage_directory}/snapshots/file_name_1_0_grid_overlay.png",
        "xtalsnapshotfullpath2": f"{storage_directory}/snapshots/file_name_1_0_outer_overlay.png",
        "xtalsnapshotfullpath3": f"{storage_directory}/snapshots/file_name_1_0.png",
        "synchrotronmode": "User",
        "undulatorgap1": 1.11,
        "filetemplate": "file_name_1_master.h5",
        "numberofimages": 20 * 6,
    }
    compare_comment(
        fetch_datacollection_attribute,
        ispyb_ids.data_collection_ids[0],
        "MX-Bluesky: Xray centring 1 - Diffraction grid scan of 20 by 6 "
        "images in 20.0 um by 20.0 um steps. Top left (px): [130,130], "
        "bottom right (px): [626,278]. Aperture: Small. ",
    )
    compare_actual_and_expected(
        ispyb_ids.data_collection_ids[0],
        dc_expected_values,
        fetch_datacollection_attribute,
        DATA_COLLECTION_COLUMN_MAP,
    )
    gridinfo_expected_values = {
        "gridInfoId": ispyb_ids.grid_ids[0],
        "dx_mm": 0.02,
        "dy_mm": 0.02,
        "steps_x": 20,
        "steps_y": 6,
        "snapshot_offsetXPixel": 130,
        "snapshot_offsetYPixel": 130,
        "orientation": "horizontal",
        "snaked": True,
        "dataCollectionId": ispyb_ids.data_collection_ids[0],
        "micronsPerPixelX": 0.806,
        "micronsPerPixelY": 0.806,
    }

    compare_actual_and_expected(
        ispyb_ids.grid_ids[0],
        gridinfo_expected_values,
        fetch_datacollection_grid_attribute,
        GRID_INFO_COLUMN_MAP,
    )
    position_id = fetch_datacollection_attribute(
        ispyb_ids.data_collection_ids[0], DATA_COLLECTION_COLUMN_MAP["positionid"]
    )
    assert position_id is None
    dc_expected_values.update(
        {
            "axisstart": 90.0,
            "axisend": 90.0,
            "datacollectionnumber": 2,
            "omegastart": 90.0,
            "filetemplate": "file_name_2_master.h5",
            "xtalsnapshotfullpath1": f"{storage_directory}/snapshots/file_name_1_90_grid_overlay.png",
            "xtalsnapshotfullpath2": f"{storage_directory}/snapshots/file_name_1_90_outer_overlay.png",
            "xtalsnapshotfullpath3": f"{storage_directory}/snapshots/file_name_1_90.png",
            "numberofimages": 20 * 6,
        }
    )
    compare_actual_and_expected(
        ispyb_ids.data_collection_ids[1],
        dc_expected_values,
        fetch_datacollection_attribute,
        DATA_COLLECTION_COLUMN_MAP,
    )
    compare_comment(
        fetch_datacollection_attribute,
        ispyb_ids.data_collection_ids[1],
        "MX-Bluesky: Xray centring 2 - Diffraction grid scan of 20 by 6 "
        "images in 20.0 um by 20.0 um steps. Top left (px): [130,130], "
        "bottom right (px): [626,278]. Aperture: Small. ",
    )
    position_id = fetch_datacollection_attribute(
        ispyb_ids.data_collection_ids[1], DATA_COLLECTION_COLUMN_MAP["positionid"]
    )
    assert position_id is None
    gridinfo_expected_values.update(
        {
            "gridInfoId": ispyb_ids.grid_ids[1],
            "steps_y": 6.0,
            "snapshot_offsetYPixel": 130.0,
            "dataCollectionId": ispyb_ids.data_collection_ids[1],
        }
    )
    compare_actual_and_expected(
        ispyb_ids.grid_ids[1],
        gridinfo_expected_values,
        fetch_datacollection_grid_attribute,
        GRID_INFO_COLUMN_MAP,
    )


def rotation_scan(
    composite: RotationScanComposite,
    parameters: RotationScan,
    oav_params: OAVParameters | None = None,
) -> MsgGenerator:
    yield from rotation_scan_internal(composite, parameters, oav_params)


@pytest.mark.system_test
def test_ispyb_deposition_in_rotation_plan(
    composite_for_rotation_scan: RotationScanComposite,
    params_for_rotation_scan: RotationScan,
    oav_parameters_for_rotation: OAVParameters,
    run_engine: RunEngine,
    fetch_comment: Callable[..., Any],
    fetch_datacollection_attribute: Callable[..., Any],
    fetch_datacollection_position_attribute: Callable[..., Any],
    tmp_path,
):
    ispyb_cb = RotationISPyBCallback()
    run_engine.subscribe(ispyb_cb)

    run_engine(
        rotation_scan(
            composite_for_rotation_scan,
            params_for_rotation_scan,
            oav_parameters_for_rotation,
        )
    )

    dcid = ispyb_cb.ispyb_ids.data_collection_ids[0]
    assert dcid is not None
    assert (
        fetch_comment(dcid) == "test Sample position (µm): (1, 2, 3) Aperture: Small. "
    )

    expected_values = replace_all_tmp_paths(
        EXPECTED_DATACOLLECTION_FOR_ROTATION
        | {
            "xtalSnapshotFullPath1": "regex:{tmp_data}/123456/snapshots/\\d{8}_oav_snapshot_0"
            ".png",
            "xtalSnapshotFullPath2": "regex:{tmp_data}/123456/snapshots/\\d{8}_oav_snapshot_90"
            ".png",
            "xtalSnapshotFullPath3": "regex:{tmp_data}/123456/snapshots/\\d{8}_oav_snapshot_180"
            ".png",
            "xtalSnapshotFullPath4": "regex:{tmp_data}/123456/snapshots/\\d{8}_oav_snapshot_270"
            ".png",
        },
        tmp_path,
    )

    compare_actual_and_expected(dcid, expected_values, fetch_datacollection_attribute)

    position_id = fetch_datacollection_attribute(
        dcid, DATA_COLLECTION_COLUMN_MAP["positionid"]
    )
    expected_values = {"posX": 0.001, "posY": 0.002, "posZ": 0.003}
    compare_actual_and_expected(
        position_id, expected_values, fetch_datacollection_position_attribute
    )


def generate_scan_data_infos(
    dummy_params,
    dummy_scan_data_info_for_begin: ScanDataInfo,
    experiment_type: IspybExperimentType,
    ispyb_ids: IspybIds,
) -> Sequence[ScanDataInfo]:
    xy_scan_data_info = scan_xy_data_info_for_update(
        ispyb_ids.data_collection_group_id, dummy_params, dummy_scan_data_info_for_begin
    )
    xy_scan_data_info.data_collection_id = ispyb_ids.data_collection_ids[0]
    if experiment_type == IspybExperimentType.GRIDSCAN_3D:
        scan_data_infos = scan_data_infos_for_update_3d(
            ispyb_ids, xy_scan_data_info, dummy_params
        )
    else:
        scan_data_infos = [xy_scan_data_info]
    return scan_data_infos



from dodal.plan_stubs.wrapped import (
    move,
    move_relative,
    set_absolute,
    set_relative,
    sleep,
    wait,
)

from mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan import (
    i04_default_grid_detect_and_xray_centre,
)
from mx_bluesky.beamlines.i04.oav_centering_plans.oav_imaging import (
    find_and_set_beam_centre_at_current_zoom_and_transmission,
    find_beam_centres,
    optimise_transmission_with_oav,
    take_oav_image_with_scintillator_in,
)
from mx_bluesky.beamlines.i04.thawing_plan import (
    thaw,
    thaw_and_murko_centre,
    thaw_and_stream_to_redis,
)

__all__ = [
    "thaw",
    "thaw_and_stream_to_redis",
    "i04_default_grid_detect_and_xray_centre",
    "thaw_and_murko_centre",
    "take_oav_image_with_scintillator_in",
    "optimise_transmission_with_oav",
    "find_beam_centres",
    "find_and_set_beam_centre_at_current_zoom_and_transmission",
    "move",
    "move_relative",
    "set_relative",
    "set_absolute",
    "sleep",
    "wait",
]



from unittest.mock import MagicMock, patch

import numpy
import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.smargon import CombinedMove, Smargon, StubPosition

from mx_bluesky.common.experiment_plans.change_aperture_then_move_plan import (
    change_aperture,
    get_results_then_change_aperture_and_move_to_xtal,
    move_to_xtal,
)
from mx_bluesky.common.parameters.device_composites import (
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.common.utils.xrc_result import XRayCentreEventHandler, XRayCentreResult


@pytest.fixture
def simple_flyscan_hit():
    return XRayCentreResult(
        centre_of_mass_mm=numpy.array([0.1, 0.2, 0.3]),
        bounding_box_mm=(
            numpy.array([0.09, 0.19, 0.29]),
            numpy.array([0.11, 0.21, 0.31]),
        ),
        max_count=20,
        total_count=57,
        sample_id=12345,
    )


@pytest.mark.parametrize("set_stub_offsets", [True, False])
def test_change_aperture_then_move_to_xtal_plans_happy_path(
    sim_run_engine: RunEngineSimulator,
    simple_flyscan_hit: XRayCentreResult,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    set_stub_offsets: bool,
):
    msgs = sim_run_engine.simulate_plan(
        change_aperture(
            simple_flyscan_hit,
            aperture_scatterguard,
        )
    )
    msgs += sim_run_engine.simulate_plan(
        move_to_xtal(simple_flyscan_hit, smargon, set_stub_offsets)
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is aperture_scatterguard.selected_aperture
        and msg.args[0] == ApertureValue.MEDIUM,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is smargon
        and msg.args[0] == CombinedMove(x=0.1, y=0.2, z=0.3),
    )
    if set_stub_offsets:
        assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj is smargon.stub_offsets
            and msg.args[0] == StubPosition.CURRENT_AS_CENTER,
        )
    else:
        assert all(
            not (msg.command == "set" and msg.obj is smargon.stub_offsets)
            for msg in msgs
        )


@patch(
    "mx_bluesky.common.experiment_plans.change_aperture_then_move_plan.change_aperture"
)
@patch("mx_bluesky.common.experiment_plans.change_aperture_then_move_plan.move_to_xtal")
@patch(
    "mx_bluesky.common.experiment_plans.change_aperture_then_move_plan.fetch_xrc_results_from_zocalo"
)
def test_get_results_then_change_aperture_and_move_to_xtal_calls_expected_plans(
    mock_fetch_results_from_zocalo: MagicMock,
    mock_change_aperture: MagicMock,
    mock__get_xrc_results: MagicMock,
    run_engine: RunEngine,
    grid_detect_xrc_devices: GridDetectThenXRayCentreComposite,
    test_fgs_params: SpecifiedThreeDGridScan,
):
    mock_flyscan_event_handler = MagicMock(spec=XRayCentreEventHandler)
    mock_flyscan_event_handler.xray_centre_results = [0]

    run_engine(
        get_results_then_change_aperture_and_move_to_xtal(
            grid_detect_xrc_devices, test_fgs_params, mock_flyscan_event_handler
        )
    )
    mock__get_xrc_results.assert_called_once()
    mock_change_aperture.assert_called_once()
    mock_fetch_results_from_zocalo.assert_called_once()



from mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.do_darks import (
    do_non_pedestal_darks,
    do_pedestal_darks,
)
from mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan import (
    single_rotation_plan,
)

__all__ = [
    "do_pedestal_darks",
    "do_non_pedestal_darks",
    "single_rotation_plan",
]



from unittest.mock import MagicMock, patch

from bluesky.plan_stubs import null
from bluesky.run_engine import RunEngine
from dodal.devices.aithre_lasershaping.goniometer import Goniometer
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection

from mx_bluesky.beamlines.aithre_lasershaping.pin_tip_centring import (
    aithre_pin_tip_centre,
)


def return_pixel(pixel, *args):
    yield from null()
    return pixel


@patch(
    "mx_bluesky.beamlines.aithre_lasershaping.pin_tip_centring.pin_tip_centre_plan",
    autospec=True,
)
async def test_when_aithre_pin_tip_centre_called_then_expected_plans_called(
    mock_pin_tip_centring_plan,
    aithre_gonio: Goniometer,
    oav: OAV,
    test_config_files: dict[str, str],
    run_engine: RunEngine,
):
    mock_pin_tip_detection = MagicMock(spec=PinTipDetection)
    run_engine(
        aithre_pin_tip_centre(
            oav,
            aithre_gonio,
            mock_pin_tip_detection,
            50,
            test_config_files["oav_config_json"],
        )
    )

    mock_pin_tip_centring_plan.assert_called_once()



from unittest.mock import MagicMock, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import Msg, MsgGenerator
from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.backlight import InOut
from dodal.devices.eiger import EigerDetector
from dodal.devices.smargon import CombinedMove
from dodal.devices.xbpm_feedback import Pause

from mx_bluesky.common.experiment_plans.inner_plans.do_fgs import ZOCALO_STAGE_GROUP
from mx_bluesky.common.parameters.constants import OavConstants
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan import (
    create_parameters_for_grid_detection,
    pin_centre_then_gridscan_plan,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionGridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import (
    PinTipCentreThenXrayCentre,
)
from tests.unit_tests.beamlines.i24.serial.conftest import fake_generator

from ...conftest import raw_params_from_file


def pin_tip_centre_then_xray_centre(
    composite: HyperionGridDetectThenXRayCentreComposite,
    parameters: PinTipCentreThenXrayCentre,
    oav_config_file: str = OavConstants.OAV_CONFIG_JSON,
) -> MsgGenerator:
    eiger: EigerDetector = composite.eiger

    eiger.set_detector_parameters(parameters.detector_params)

    def pin_centre_flyscan_then_fetch_results() -> MsgGenerator:
        yield from pin_centre_then_gridscan_plan(composite, parameters, oav_config_file)

    yield from pin_centre_flyscan_then_fetch_results()


@pytest.fixture
def test_grid_params():
    return {
        "transmission_frac": 1.0,
        "exposure_time_s": 0,
        "x_start_um": 0,
        "y_start_um": 0,
        "y2_start_um": 0,
        "z_start_um": 0,
        "z2_start_um": 0,
        "x_steps": 10,
        "y_steps": 10,
        "z_steps": 10,
        "x_step_size_um": 0.1,
        "y_step_size_um": 0.1,
        "z_step_size_um": 0.1,
    }


@pytest.fixture
def test_pin_centre_then_xray_centre_params(
    tmp_path,
) -> PinTipCentreThenXrayCentre:
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_pin_centre_then_xray_centre_parameters.json",
        tmp_path,
    )
    params = PinTipCentreThenXrayCentre(**params)
    return params


@pytest.fixture
def pin_centre_then_xray_centre_params_with_patched_create_params(
    test_fgs_params: SpecifiedThreeDGridScan,
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
):
    with patch(
        "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.create_parameters_for_grid_detection"
    ) as mock_create_params:
        test_pin_centre_then_xray_centre_params.set_specified_grid_params(
            test_fgs_params
        )
        mock_create_params.return_value = test_pin_centre_then_xray_centre_params
        yield test_pin_centre_then_xray_centre_params


def test_when_create_parameters_for_grid_detection_then_parameters_created(
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
):
    grid_detect_params = create_parameters_for_grid_detection(
        test_pin_centre_then_xray_centre_params
    )

    assert grid_detect_params.exposure_time_s == 0.1


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.pin_tip_centre_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.detect_grid_and_do_gridscan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.fetch_xrc_results_from_zocalo",
    new=MagicMock(),
)
def test_when_pin_centre_xray_centre_called_then_plan_runs_correctly(
    mock_detect_and_do_gridscan: MagicMock,
    mock_pin_tip_centre: MagicMock,
    pin_centre_then_xray_centre_params_with_patched_create_params: PinTipCentreThenXrayCentre,
    hyperion_grid_detect_xrc_devices: HyperionGridDetectThenXRayCentreComposite,
    test_config_files,
    run_engine: RunEngine,
):
    run_engine(
        pin_centre_then_gridscan_plan(
            hyperion_grid_detect_xrc_devices,
            pin_centre_then_xray_centre_params_with_patched_create_params,
            test_config_files["oav_config_json"],
        )
    )

    mock_detect_and_do_gridscan.assert_called_once()
    mock_pin_tip_centre.assert_called_once()


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.pin_tip_centre_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.detect_grid_and_do_gridscan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.fetch_xrc_results_from_zocalo",
    new=MagicMock(),
)
def test_pin_centre_then_gridscan_plan_activates_ispyb_callback_before_pin_tip_centre_plan(
    mock_detect_grid_and_do_gridscan,
    mock_pin_tip_centre_plan,
    sim_run_engine: RunEngineSimulator,
    pin_centre_then_xray_centre_params_with_patched_create_params: PinTipCentreThenXrayCentre,
    hyperion_grid_detect_xrc_devices: HyperionGridDetectThenXRayCentreComposite,
    test_config_files,
):
    mock_detect_grid_and_do_gridscan.return_value = iter(
        [Msg("detect_grid_and_do_gridscan")]
    )
    mock_pin_tip_centre_plan.return_value = iter([Msg("pin_tip_centre_plan")])

    msgs = sim_run_engine.simulate_plan(
        pin_centre_then_gridscan_plan(
            hyperion_grid_detect_xrc_devices,
            pin_centre_then_xray_centre_params_with_patched_create_params,
            test_config_files["oav_config_json"],
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "open_run"
        and "GridscanISPyBCallback" in msg.kwargs["activate_callbacks"],
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "pin_tip_centre_plan"
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "detect_grid_and_do_gridscan"
    )
    assert_message_and_return_remaining(msgs, lambda msg: msg.command == "close_run")


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.pin_tip_centre_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.detect_grid_and_do_gridscan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.fetch_xrc_results_from_zocalo",
    new=MagicMock(),
)
def test_pin_centre_then_gridscan_plan_sets_up_backlight_and_aperture(
    mock_detect_grid_and_do_gridscan,
    mock_pin_tip_centre_plan,
    hyperion_grid_detect_xrc_devices: HyperionGridDetectThenXRayCentreComposite,
    sim_run_engine: RunEngineSimulator,
    pin_centre_then_xray_centre_params_with_patched_create_params: PinTipCentreThenXrayCentre,
    test_config_files,
):
    mock_detect_grid_and_do_gridscan.return_value = iter(
        [Msg("detect_grid_and_do_gridscan")]
    )
    mock_pin_tip_centre_plan.return_value = iter([Msg("pin_tip_centre_plan")])

    msgs = sim_run_engine.simulate_plan(
        pin_centre_then_gridscan_plan(
            hyperion_grid_detect_xrc_devices,
            pin_centre_then_xray_centre_params_with_patched_create_params,
            test_config_files["oav_config_json"],
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "backlight"
        and msg.args == (InOut.IN,)
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj
        == hyperion_grid_detect_xrc_devices.aperture_scatterguard.selected_aperture
        and msg.args == (ApertureValue.OUT_OF_BEAM,)
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )

    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "pin_tip_centre_plan"
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.pin_tip_centre_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.detect_grid_and_do_gridscan",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.fetch_xrc_results_from_zocalo",
    new=MagicMock(),
)
def test_pin_centre_then_gridscan_plan_goes_to_the_starting_chi_and_phi(
    mock_detect_grid_and_do_gridscan,
    mock_pin_tip_centre_plan,
    sim_run_engine: RunEngineSimulator,
    pin_centre_then_xray_centre_params_with_patched_create_params: PinTipCentreThenXrayCentre,
    test_config_files,
    hyperion_grid_detect_xrc_devices,
):
    params = pin_centre_then_xray_centre_params_with_patched_create_params
    mock_detect_grid_and_do_gridscan.return_value = iter(
        [Msg("detect_grid_and_do_gridscan")]
    )
    mock_pin_tip_centre_plan.return_value = iter([Msg("pin_tip_centre_plan")])

    params.phi_start_deg = (expected_phi := 30)
    params.chi_start_deg = (expected_chi := 50)

    msgs = sim_run_engine.simulate_plan(
        pin_centre_then_gridscan_plan(
            hyperion_grid_detect_xrc_devices,
            params,
            test_config_files["oav_config_json"],
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "gonio"
        and msg.args[0] == CombinedMove(phi=expected_phi, chi=expected_chi, omega=None)
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )

    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "pin_tip_centre_plan"
    )


@pytest.mark.parametrize("transmission_frac", [1, 0.5, 0.25])
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.pin_tip_centre_plan"
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan"
)
@patch(
    "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.fetch_xrc_results_from_zocalo",
    new=MagicMock(),
)
def test_pin_tip_centre_then_xray_centre_sets_transmission_fraction_and_xbpm_is_paused_and_both_reverted(
    mock_run_gridscan: MagicMock,
    mock_grid_detection_plan: MagicMock,
    mock_pin_tip_centre_plan: MagicMock,
    mock_grid_detection_callback: MagicMock,
    test_grid_params,
    transmission_frac: float,
    sim_run_engine: RunEngineSimulator,
    hyperion_grid_detect_xrc_devices: HyperionGridDetectThenXRayCentreComposite,
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
):
    mock_grid_detection_callback.return_value.get_grid_parameters.return_value = (
        test_grid_params
    )

    test_pin_centre_then_xray_centre_params.transmission_frac = transmission_frac

    msgs = sim_run_engine.simulate_plan(
        pin_tip_centre_then_xray_centre(
            hyperion_grid_detect_xrc_devices,
            test_pin_centre_then_xray_centre_params,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "set"
        and msg.obj.name == "xbpm_feedback-pause_feedback"
        and msg.args[0] == Pause.PAUSE,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "set"
        and msg.obj.name == "attenuator"
        and msg.args[0] == transmission_frac,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "set"
        and msg.obj.name == "xbpm_feedback-pause_feedback"
        and msg.args[0] == Pause.RUN,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "set"
        and msg.obj.name == "attenuator"
        and msg.args[0] == 1.0,
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.pin_tip_centre_plan",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.detect_grid_and_do_gridscan",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.fetch_xrc_results_from_zocalo"
)
def test_pin_centre_then_xrc_stages_and_unstages_zocalo_and_gets_results(
    mock_fetch_results_and_move: MagicMock,
    hyperion_grid_detect_xrc_devices: HyperionGridDetectThenXRayCentreComposite,
    sim_run_engine: RunEngineSimulator,
    pin_centre_then_xray_centre_params_with_patched_create_params: PinTipCentreThenXrayCentre,
):
    msgs = sim_run_engine.simulate_plan(
        pin_tip_centre_then_xray_centre(
            hyperion_grid_detect_xrc_devices,
            pin_centre_then_xray_centre_params_with_patched_create_params,
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "stage"
        and msg.obj.name == "zocalo"
        and msg.kwargs["group"] == ZOCALO_STAGE_GROUP,
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "unstage" and msg.obj.name == "zocalo",
    )
    mock_fetch_results_and_move.assert_called_once()


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.pin_tip_centre_plan",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
    lambda *_: fake_generator("_"),
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.common_flyscan_xray_centre",
    lambda *_: fake_generator("_"),
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
    new=MagicMock(),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_centre_then_gridscan_plan.fetch_xrc_results_from_zocalo"
)
def test_detect_grid_and_do_gridscan_gives_params_specified_grid(
    mock_fetch_xrc_results: MagicMock,
    mock_create_flyscan_params: MagicMock,
    test_pin_centre_then_xray_centre_params: PinTipCentreThenXrayCentre,
    hyperion_grid_detect_xrc_devices: HyperionGridDetectThenXRayCentreComposite,
    test_fgs_params: SpecifiedThreeDGridScan,
    test_config_files,
    run_engine: RunEngine,
):
    mock_create_flyscan_params.return_value = test_fgs_params
    run_engine(
        pin_centre_then_gridscan_plan(
            hyperion_grid_detect_xrc_devices,
            test_pin_centre_then_xray_centre_params,
            test_config_files["oav_config_json"],
        )
    )
    mock_fetch_xrc_results.assert_called_once()
    assert mock_fetch_xrc_results.call_args[0][1] == test_fgs_params



from blueapi.client.event_bus import AnyEvent
from blueapi.core import DataEvent
from bluesky.run_engine import Dispatcher
from bluesky_stomp.messaging import MessageContext, StompClient
from bluesky_stomp.models import MessageTopic
from event_model import DocumentNames

from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER as LOGGER

BLUEAPI_EVENT_TOPIC = "public.worker.event"


class StompDispatcher(Dispatcher):
    def __init__(self, stomp_client: StompClient):
        super().__init__()
        self._client = stomp_client

    def __enter__(self):
        self._subscription_id = self._client.subscribe(
            MessageTopic(name=BLUEAPI_EVENT_TOPIC), self._on_event
        )
        LOGGER.info("Connecting to stomp broker...")
        self._client.connect()

    def __exit__(self, exc_type, exc_val, exc_tb):
        LOGGER.info("Disconnecting from stomp and unsubscribing...")
        self._client.disconnect()
        self._client.unsubscribe(self._subscription_id)

    def _on_event(self, event: AnyEvent, context: MessageContext):
        match event:
            case DataEvent(name=name, doc=doc):  # type: ignore
                self.process(DocumentNames[name], doc)



import bluesky.preprocessors as bpp
from bluesky import plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from ophyd_async.fastcs.jungfrau import (
    AcquisitionType,
    GainMode,
    create_jungfrau_internal_triggering_info,
    create_jungfrau_pedestal_triggering_info,
)
from pydantic import PositiveInt

from mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils import (
    fly_jungfrau,
)
from mx_bluesky.common.utils.log import LOGGER

PEDESTAL_DARKS_RUN = "PEDESTAL DARKS RUN"
STANDARD_DARKS_RUN = "STANDARD DARKS RUN"


def do_pedestal_darks(
    exp_time_s: float = 0.001,
    pedestal_frames: PositiveInt = 20,
    pedestal_loops: PositiveInt = 200,
    filename: str = "pedestal_darks",
    jungfrau: CommissioningJungfrau = inject("jungfrau"),
) -> MsgGenerator:
    """Acquire darks in pedestal mode, using dynamic gain mode. This calibrates the offsets
    for the jungfrau, and must be performed before acquiring real data in dynamic gain mode.

    When Bluesky triggers the detector in pedestal mode, with pedestal frames F and pedestal loops L,
    the acquisition is managed at the driver level to:
    1. Acquire F-1 frames in dynamic gain mode
    2. Acquire 1 frame in ForceSwitchG1 gain mode
    3. Repeat steps 1-2 L times
    4. Do the first three steps a second time, except use ForceSwitchG2 instead of ForceSwitchG1
    during step 2.

    A pedestal scan should be acquired when detector configuration and environmental conditions change, but due to small
    in instabilities in beamline conditions, it is recommended to run a pedestal scan on roughly an hourly basis.

    Args:
        exp_time_s: Length of detector exposure for each frame.
        pedestal_frames: Number of frames acquired per pedestal loop.
        pedestal_loops: Number of times to acquire a set of pedestal_frames
        filename: Name of output file
        jungfrau: Jungfrau device
    """

    @bpp.set_run_key_decorator(PEDESTAL_DARKS_RUN)
    @bpp.run_decorator(
        md={
            "subplan_name": PEDESTAL_DARKS_RUN,
            "detector_file_template": filename,
        }
    )
    @bpp.stage_decorator([jungfrau])
    def _do_decorated_plan():
        trigger_info = create_jungfrau_pedestal_triggering_info(
            exp_time_s, pedestal_frames, pedestal_loops
        )
        LOGGER.info(
            "Jungfrau will be triggered in pedestal mode and in dynamic gain mode"
        )
        yield from bps.mv(
            jungfrau.drv.acquisition_type,
            AcquisitionType.PEDESTAL,
        )
        yield from fly_jungfrau(
            jungfrau,
            trigger_info,
            GainMode.DYNAMIC,
            wait=True,
            log_on_percentage_prefix="Jungfrau pedestal dynamic gain mode darks triggers received",
        )

    yield from _do_decorated_plan()


def do_non_pedestal_darks(
    gain_mode: GainMode,
    exp_time_s: float = 0.001,
    total_triggers: PositiveInt = 1000,
    filename: str = "darks",
    jungfrau: CommissioningJungfrau = inject("jungfrau"),
) -> MsgGenerator:
    """Internally take a set of images at a given gain mode.

    Non-pedestal darks are useful for detector panel cross-checks and for calculating masks.

    Args:
        gain_mode: Which gain mode to put the Jungfrau into before starting the acquisition.
        exp_time_s: Length of detector exposure for each trigger.
        total_triggers: Total triggers for the dark scan.
        jungfrau: Jungfrau device
        filename: Name of output file
    """

    @bpp.set_run_key_decorator(STANDARD_DARKS_RUN)
    @bpp.run_decorator(
        md={
            "subplan_name": STANDARD_DARKS_RUN,
            "detector_file_template": filename,
        }
    )
    @bpp.stage_decorator([jungfrau])
    def _do_decorated_plan():
        trigger_info = create_jungfrau_internal_triggering_info(
            total_triggers, exp_time_s
        )

        yield from fly_jungfrau(
            jungfrau,
            trigger_info,
            gain_mode,
            wait=True,
            log_on_percentage_prefix=f"Jungfrau {gain_mode} gain mode darks triggers received",
        )

    yield from _do_decorated_plan()



import asyncio
from functools import partial
from unittest.mock import AsyncMock, MagicMock, patch

import bluesky.plan_stubs as bps
from bluesky.preprocessors import run_decorator
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.beamlines.i24 import CommissioningJungfrau
from ophyd_async.core import set_mock_value
from ophyd_async.fastcs.jungfrau import GainMode

from mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.do_external_acquisition import (
    do_external_acquisition,
)
from mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils import (
    JF_COMPLETE_GROUP,
)


def test_full_do_external_acquisition(
    jungfrau: CommissioningJungfrau, run_engine: RunEngine, caplog
):
    @run_decorator(
        md={
            "detector_file_template": "test",
        }
    )
    def test_plan():
        status = yield from do_external_acquisition(
            0.001, GainMode.DYNAMIC, 5, jungfrau=jungfrau
        )
        assert not status.done
        val = 0
        while not status.done:
            val += 1
            set_mock_value(jungfrau._writer.frame_counter, val)

            # Let status update
            yield from bps.wait_for([partial(asyncio.sleep, 0)])
        yield from bps.wait(JF_COMPLETE_GROUP)

    jungfrau._controller.arm = AsyncMock()
    run_engine(test_plan())
    for i in range(20, 120, 20):
        assert f"Jungfrau data collection triggers received: {i}%" in caplog.messages


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils.log_on_percentage_complete"
)
def test_do_external_acquisition_does_wait(
    mock_log_on_percent_complete: MagicMock,
    sim_run_engine: RunEngineSimulator,
    jungfrau: CommissioningJungfrau,
):
    msgs = sim_run_engine.simulate_plan(
        do_external_acquisition(0.01, GainMode.DYNAMIC, 1, wait=True, jungfrau=jungfrau)
    )
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait" and msg.kwargs["group"] == JF_COMPLETE_GROUP,
    )



import subprocess
import time
from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from os import environ, getcwd
from pathlib import Path
from threading import Event
from time import sleep
from unittest.mock import MagicMock, call, patch

import pytest
from blueapi.client.event_bus import AnyEvent, EventBusClient
from blueapi.config import ApplicationConfig, ConfigLoader
from blueapi.core import BlueskyContext, DataEvent
from blueapi.service.model import TaskRequest
from blueapi.worker import WorkerEvent, WorkerState
from bluesky import RunEngine, RunEngineInterrupted
from bluesky import plan_stubs as bps
from bluesky_stomp.messaging import MessageContext

from mx_bluesky.common.external_interaction.alerting import Metadata
from mx_bluesky.common.parameters.components import get_param_version
from mx_bluesky.common.parameters.constants import Status
from mx_bluesky.hyperion._plan_runner_params import RobotUnload
from mx_bluesky.hyperion.blueapi.parameters import LoadCentreCollectParams
from mx_bluesky.hyperion.plan_runner import PlanError
from mx_bluesky.hyperion.supervisor import SupervisorRunner

from ....conftest import (
    TEST_CONTAINER,
    TEST_SAMPLE_ID,
    TEST_VISIT,
    raw_params_from_file,
)
from .dummy_plans import (
    BEAMLINE_ERROR_SAMPLE_ID,
    CRYSTAL_NOT_FOUND_SAMPLE_ID,
    WaitForFeedbackParams,
)

BLUEAPI_SERVER_CONFIG = (
    "tests/system_tests/hyperion/supervisor/system_test_blueapi.yaml"
)


@pytest.fixture(scope="module")
def tpe():
    return ThreadPoolExecutor(max_workers=1)


@pytest.fixture(scope="module")
def mock_blueapi_server():
    with subprocess.Popen(
        [
            "blueapi",
            "--config",
            BLUEAPI_SERVER_CONFIG,
            "serve",
        ],
        env=environ | {"PYTHONPATH": getcwd() + "/tests"},
    ) as blueapi_server:
        try:
            yield blueapi_server
        finally:
            blueapi_server.terminate()


@pytest.fixture
def mock_bluesky_context(run_engine: RunEngine):
    loader = ConfigLoader(ApplicationConfig)
    loader.use_values_from_yaml(
        Path("tests/system_tests/hyperion/supervisor/supervisor_config.yaml")
    )
    supervisor_config = loader.load()
    yield BlueskyContext(configuration=supervisor_config, run_engine=run_engine)


@pytest.fixture
def client_config() -> ApplicationConfig:
    loader = ConfigLoader(ApplicationConfig)
    loader.use_values_from_yaml(
        Path("tests/system_tests/hyperion/supervisor/client_config.yaml")
    )
    return loader.load()


def get_event_bus_client(supervisor: SupervisorRunner) -> EventBusClient:
    return supervisor.blueapi_client._events  # type: ignore


@pytest.fixture
def supervisor_runner(supervisor_runner_no_ping: SupervisorRunner):
    supervisor_runner_no_ping.reset_callback_watchdog_timer()
    yield supervisor_runner_no_ping


@pytest.fixture
def supervisor_runner_no_ping(
    mock_bluesky_context: BlueskyContext,
    client_config: ApplicationConfig,
    mock_blueapi_server,
    monkeypatch,
):
    monkeypatch.setenv("BEAMLINE", "i03")
    runner = SupervisorRunner(mock_bluesky_context, client_config, True)
    timeout = time.monotonic() + 30
    while time.monotonic() < timeout:
        if runner.is_connected():
            return runner
        sleep(1)
    else:
        raise AssertionError("Failed to connect to blueapi")


def handle_event(plan_started: Event, event_payload: AnyEvent, context: MessageContext):
    match event_payload:
        case DataEvent() as data_event:
            if (
                data_event.name == "start"
                and data_event.doc["plan_name"] == "robot_unload"
            ):
                plan_started.set()


@pytest.mark.system_test
def test_supervisor_connects_to_blueapi_and_stomp(
    mock_blueapi_server,
    mock_bluesky_context: BlueskyContext,
    client_config: ApplicationConfig,
    supervisor_runner: SupervisorRunner,
):
    params = RobotUnload.model_validate(
        {"parameter_model_version": get_param_version()}
    )
    ebc = get_event_bus_client(supervisor_runner)

    received_message_event = Event()

    ebc.subscribe_to_all_events(partial(handle_event, received_message_event))
    assert supervisor_runner.current_status == Status.IDLE
    supervisor_runner.run_engine(
        supervisor_runner.decode_and_execute(TEST_VISIT, [params])
    )
    received_message_event.wait()


@pytest.mark.skip(reason="https://github.com/DiamondLightSource/blueapi/issues/1312")
def test_supervisor_continues_to_next_instruction_on_warning_error(
    supervisor_runner: SupervisorRunner,
):
    params = RobotUnload.model_validate(
        {"parameter_model_version": get_param_version()}
    )
    supervisor_runner.run_engine(
        supervisor_runner.decode_and_execute("raise_warning_error", [params])
    )
    assert supervisor_runner.current_status == Status.FAILED


def test_supervisor_raises_request_abort_when_shutdown_requested(
    supervisor_runner: SupervisorRunner, tpe: ThreadPoolExecutor
):
    params = RobotUnload.model_validate(
        {"parameter_model_version": get_param_version()}
    )
    ebc = get_event_bus_client(supervisor_runner)
    plan_aborted = Event()
    plan_called = Event()

    def handle_abort(event_payload: AnyEvent, context: MessageContext):
        match event_payload:
            case WorkerEvent() as worker_event:
                if worker_event.state == WorkerState.ABORTING:
                    plan_aborted.set()

    ebc.subscribe_to_all_events(partial(handle_event, plan_called))
    ebc.subscribe_to_all_events(handle_abort)

    def shutdown_in_background():
        plan_called.wait(10)
        assert supervisor_runner.current_status == Status.BUSY
        assert supervisor_runner.blueapi_client.state == WorkerState.RUNNING
        supervisor_runner.shutdown()
        assert supervisor_runner.current_status == Status.ABORTING

    fut = tpe.submit(shutdown_in_background)

    def execute_remotely():
        yield from supervisor_runner.decode_and_execute("wait_for_abort", [params])
        # Simulate supervisor going round in a loop, since async abort request may not be processed before requested execution
        # completes

        yield from bps.sleep(10)

    with pytest.raises(RunEngineInterrupted):
        supervisor_runner.run_engine(execute_remotely())

    assert supervisor_runner.blueapi_client.state == WorkerState.IDLE
    assert plan_aborted.wait(10)
    fut.result()


def test_supervisor_raises_plan_error_when_plan_fails_with_other_exception(
    supervisor_runner: SupervisorRunner,
):
    params = RobotUnload.model_validate(
        {"parameter_model_version": get_param_version()}
    )
    with pytest.raises(PlanError, match="Exception raised during plan execution:"):
        supervisor_runner.run_engine(
            supervisor_runner.decode_and_execute("raise_other_error", [params])
        )
    assert supervisor_runner.current_status == Status.FAILED


def test_shutdown_raises_run_engine_interrupted_when_idle(
    supervisor_runner: SupervisorRunner, tpe: ThreadPoolExecutor
):
    plan_started = Event()

    def idle_plan():
        plan_started.set()
        while True:
            yield from bps.sleep(1)

    def shutdown_in_background():
        plan_started.wait(10)
        supervisor_runner.shutdown()

    fut = tpe.submit(shutdown_in_background)

    with pytest.raises(RunEngineInterrupted):
        supervisor_runner.run_engine(idle_plan())

    fut.result()


async def test_supervisor_checks_for_external_callback_ping(
    supervisor_runner_no_ping: SupervisorRunner, tpe: ThreadPoolExecutor
):
    ebc = get_event_bus_client(supervisor_runner_no_ping)
    received_message_event = Event()
    ebc.subscribe_to_all_events(partial(handle_event, received_message_event))
    params = RobotUnload.model_validate(
        {"parameter_model_version": get_param_version()}
    )

    def run_test_in_background():
        sleep(1)
        assert not received_message_event.is_set()
        supervisor_runner_no_ping.reset_callback_watchdog_timer()
        received_message_event.wait(1)

    fut = tpe.submit(run_test_in_background)
    supervisor_runner_no_ping.run_engine(
        supervisor_runner_no_ping.decode_and_execute(TEST_VISIT, [params])
    )
    fut.result()


def test_supervisor_raises_plan_error_when_external_callbacks_watchdog_expired(
    supervisor_runner_no_ping: SupervisorRunner,
):
    runner = supervisor_runner_no_ping
    runner.EXTERNAL_CALLBACK_WATCHDOG_TIMER_S = 0.5  # type: ignore
    runner.reset_callback_watchdog_timer()
    params = RobotUnload.model_validate(
        {"parameter_model_version": get_param_version()}
    )
    # Allow callback watchdog to expire
    sleep(1)
    with pytest.raises(PlanError, match="External callback watchdog timer expired.*"):
        runner.run_engine(runner.decode_and_execute(TEST_VISIT, [params]))


def test_supervisor_calls_load_centre_collect(
    supervisor_runner: SupervisorRunner, tmp_path
):
    params = LoadCentreCollectParams(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/external_load_centre_collect_params.json",
            tmp_path,
        )
    )

    supervisor_runner.run_engine(
        supervisor_runner.decode_and_execute(TEST_VISIT, [params])
    )


@patch("mx_bluesky.hyperion.supervisor._task_monitor.TaskMonitor.DEFAULT_TIMEOUT_S", 1)
@patch("mx_bluesky.hyperion.supervisor._task_monitor.get_alerting_service")
def test_supervisor_alerts_repeatedly_when_waiting_for_beam(
    mock_get_alerting_service: MagicMock,
    supervisor_runner: SupervisorRunner,
):
    mock_alerting_service = mock_get_alerting_service.return_value
    params = WaitForFeedbackParams(
        time_for_beam_stable=2.5,
        time_in_plan=0.5,
        sample_id=TEST_SAMPLE_ID,
        visit=TEST_VISIT,
        sample_puck=TEST_CONTAINER,
    )
    task_request = TaskRequest(
        name="wait_for_feedback",
        params={"parameters": params},
        instrument_session=TEST_VISIT,
    )
    supervisor_runner._run_task_remotely(task_request)

    mock_alerting_service.raise_alert.assert_has_calls(
        [
            call(
                "Hyperion is paused waiting for beam on i03.",
                "Hyperion has been paused waiting for beam for 0 minutes.",
                {
                    Metadata.SAMPLE_ID: TEST_SAMPLE_ID,
                    Metadata.VISIT: TEST_VISIT,
                    Metadata.CONTAINER: TEST_CONTAINER,
                },
            ),
            call(
                "Hyperion is paused waiting for beam on i03.",
                "Hyperion has been paused waiting for beam for 0 minutes.",
                {
                    Metadata.SAMPLE_ID: TEST_SAMPLE_ID,
                    Metadata.VISIT: TEST_VISIT,
                    Metadata.CONTAINER: TEST_CONTAINER,
                },
            ),
        ]
    )


@patch("mx_bluesky.hyperion.supervisor._task_monitor.TaskMonitor.DEFAULT_TIMEOUT_S", 1)
@patch("mx_bluesky.hyperion.supervisor._task_monitor.get_alerting_service")
def test_supervisor_alerts_with_error_and_aborts_task_when_stuck_not_waiting_for_beam(
    mock_get_alerting_service: MagicMock,
    supervisor_runner: SupervisorRunner,
):
    mock_alerting_service = mock_get_alerting_service.return_value
    params = WaitForFeedbackParams(
        time_for_beam_stable=0.5,
        time_in_plan=1.5,
        sample_id=TEST_SAMPLE_ID,
        visit=TEST_VISIT,
        sample_puck=TEST_CONTAINER,
    )

    task_request = TaskRequest(
        name="wait_for_feedback",
        params={"parameters": params},
        instrument_session=TEST_VISIT,
    )

    with pytest.raises(PlanError, match="Exception raised during plan execution"):
        supervisor_runner._run_task_remotely(task_request)

    mock_alerting_service.raise_alert.assert_has_calls(
        [
            call(
                "UDC encountered an error on i03",
                "Hyperion Supervisor detected that BlueAPI was stuck for 1 seconds.",
                {
                    Metadata.SAMPLE_ID: TEST_SAMPLE_ID,
                    Metadata.VISIT: TEST_VISIT,
                    Metadata.CONTAINER: TEST_CONTAINER,
                },
            ),
        ]
    )


@patch("mx_bluesky.hyperion.supervisor._task_monitor.TaskMonitor.DEFAULT_TIMEOUT_S", 1)
@patch("mx_bluesky.hyperion.supervisor._task_monitor.get_alerting_service")
def test_supervisor_no_alerts_when_not_stuck(
    mock_get_alerting_service: MagicMock,
    supervisor_runner: SupervisorRunner,
):
    mock_alerting_service = mock_get_alerting_service.return_value
    params = WaitForFeedbackParams(
        time_for_beam_stable=0.5,
        time_in_plan=0.5,
        sample_id=TEST_SAMPLE_ID,
        visit=TEST_VISIT,
        sample_puck=TEST_CONTAINER,
    )

    task_request = TaskRequest(
        name="wait_for_feedback",
        params={"parameters": params},
        instrument_session=TEST_VISIT,
    )

    supervisor_runner._run_task_remotely(task_request)

    mock_alerting_service.raise_alert.assert_not_called()


def test_supervisor_decode_and_execute_raises_planerror_if_blueapi_plan_raises_exception(
    supervisor_runner: SupervisorRunner, tmp_path
):
    params = LoadCentreCollectParams(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/external_load_centre_collect_params.json",
            tmp_path,
        )
    )
    params.sample_id = BEAMLINE_ERROR_SAMPLE_ID
    with pytest.raises(PlanError, match="Simulated beamline error"):
        supervisor_runner.run_engine(
            supervisor_runner.decode_and_execute(TEST_VISIT, [params])
        )


def test_supervisor_decode_and_execute_continues_if_blueapi_plan_raises_sample_error(
    supervisor_runner: SupervisorRunner, tmp_path
):
    params = LoadCentreCollectParams(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/external_load_centre_collect_params.json",
            tmp_path,
        )
    )
    params.sample_id = CRYSTAL_NOT_FOUND_SAMPLE_ID
    supervisor_runner.run_engine(
        supervisor_runner.decode_and_execute(TEST_VISIT, [params, params])
    )



import json
from unittest.mock import MagicMock, patch

import pytest
from event_model import Event

from mx_bluesky.beamlines.i04.callbacks.murko_callback import (
    MurkoCallback,
    extrapolate_omega,
)

test_oav_uuid = "UUID"
test_smargon_data = 90


def event_template(data_dict: dict, timestamp=1666604299.0) -> Event:
    return {
        "descriptor": "bd45c2e5-2b85-4280-95d7-a9a15800a78b",
        "time": 1666604299.01,
        "data": data_dict,
        "timestamps": dict.fromkeys(data_dict.keys(), timestamp),
        "seq_num": 1,
        "uid": "29033ecf-e052-43dd-98af-c7cdd62e8173",
        "filled": {},
    }


test_oav_full_screen_event = event_template(
    {
        "oav_to_redis_forwarder-uuid": test_oav_uuid,
        "oav_full_screen-microns_per_pixel_x": 1.2,
        "oav_full_screen-microns_per_pixel_y": 2.5,
        "oav_full_screen-beam_centre_i": 158,
        "oav_full_screen-beam_centre_j": 452,
    }
)

test_oav_roi_event = event_template(
    {
        "oav_to_redis_forwarder-uuid": test_oav_uuid,
        "oav-microns_per_pixel_x": 3.2,
        "oav-microns_per_pixel_y": 4.1,
        "oav-beam_centre_i": 201,
        "oav-beam_centre_j": 342,
    }
)
test_smargon_event = event_template({"gonio-omega": test_smargon_data})

test_start_document = {
    "uid": "event_uuid",
    "sample_id": 12345,
}

metadata_based_on_test_events_full_screen = {
    "sample_id": 12345,
    "microns_per_x_pixel": 1.2,
    "microns_per_y_pixel": 2.5,
    "beam_centre_i": 158,
    "beam_centre_j": 452,
    "omega_angle": test_smargon_data,
    "uuid": test_oav_uuid,
}

metadata_based_on_test_events_roi = {
    "sample_id": 12345,
    "microns_per_x_pixel": 3.2,
    "microns_per_y_pixel": 4.1,
    "beam_centre_i": 201,
    "beam_centre_j": 342,
    "omega_angle": test_smargon_data,
    "uuid": test_oav_uuid,
}


@pytest.fixture
def murko_with_mock_call(murko_callback) -> MurkoCallback:
    murko_callback.call_murko = MagicMock()
    murko_callback.murko_metadata = {}
    return murko_callback


def test_when_oav_data_arrives_then_murko_not_called(
    murko_with_mock_call: MurkoCallback,
):
    murko_with_mock_call.event(test_oav_full_screen_event)
    murko_with_mock_call.call_murko.assert_not_called()  # type: ignore


def test_when_smargon_data_arrives_with_no_image_then_murko_not_called(
    murko_with_mock_call: MurkoCallback,
):
    murko_with_mock_call.event(test_smargon_event)
    murko_with_mock_call.call_murko.assert_not_called()  # type: ignore


def test_when_smargon_data_arrives_then_image_data_then_murko_not_called(
    murko_with_mock_call: MurkoCallback,
):
    murko_with_mock_call.event(test_smargon_event)
    murko_with_mock_call.event(test_oav_full_screen_event)
    murko_with_mock_call.call_murko.assert_not_called()  # type: ignore


def test_given_image_data_when_first_two_sets_of_smargon_data_arrive_then_murko_called_with_latest_image_and_omega(
    murko_with_mock_call: MurkoCallback,
):
    murko_with_mock_call.event(test_oav_full_screen_event)
    murko_with_mock_call.event(test_smargon_event)
    murko_with_mock_call.call_murko.assert_called_once_with(  # type: ignore
        test_oav_uuid, test_smargon_data
    )

    murko_with_mock_call.call_murko.reset_mock()  # type: ignore

    murko_with_mock_call.event(event_template({"gonio-omega": (second_omega := 30)}))
    murko_with_mock_call.call_murko.assert_called_once_with(  # type: ignore
        test_oav_uuid, second_omega
    )


def test_given_two_sets_of_smargon_data_then_next_image_calls_murko_with_extrapolated_omega(
    murko_with_mock_call: MurkoCallback,
):
    murko_with_mock_call.event(test_oav_full_screen_event)
    murko_with_mock_call.event(event_template({"gonio-omega": 10}, 0))
    murko_with_mock_call.event(event_template({"gonio-omega": 15}, 5))

    murko_with_mock_call.call_murko.reset_mock()  # type:ignore

    murko_with_mock_call.event(
        event_template({"oav_to_redis_forwarder-uuid": test_oav_uuid}, 10)
    )

    murko_with_mock_call.call_murko.assert_called_once_with(test_oav_uuid, 20)  # type: ignore


def test_given_three_sets_of_smargon_data_then_next_image_calls_murko_with_extrapolated_omega_from_last_two(
    murko_with_mock_call: MurkoCallback,
):
    murko_with_mock_call.event(test_oav_full_screen_event)
    murko_with_mock_call.event(event_template({"gonio-omega": 10}, 0))
    murko_with_mock_call.event(event_template({"gonio-omega": 15}, 5))
    murko_with_mock_call.event(event_template({"gonio-omega": 17}, 6))

    murko_with_mock_call.call_murko.reset_mock()  # type:ignore

    murko_with_mock_call.event(
        event_template({"oav_to_redis_forwarder-uuid": test_oav_uuid}, 10)
    )

    murko_with_mock_call.call_murko.assert_called_once_with(test_oav_uuid, 25)  # type: ignore


def test_when_murko_called_with_event_data_then_meta_data_put_in_redis(
    murko_callback: MurkoCallback,
):
    murko_callback.start(test_start_document)  # type: ignore
    murko_callback.event(test_oav_full_screen_event)
    murko_callback.event(test_smargon_event)

    expected_metadata = metadata_based_on_test_events_full_screen

    murko_callback.redis_client.hset.assert_called_once_with(  # type: ignore
        "murko:12345:metadata", test_oav_uuid, json.dumps(expected_metadata)
    )
    murko_callback.redis_client.publish.assert_called_once_with(  # type: ignore
        "murko", json.dumps(expected_metadata)
    )


@pytest.mark.parametrize(
    "first_event, second_event, expected_md_after_first, expected_md_after_second",
    [
        (
            test_oav_full_screen_event,
            test_oav_roi_event,
            metadata_based_on_test_events_full_screen,
            metadata_based_on_test_events_roi,
        ),
        (
            test_oav_roi_event,
            test_oav_full_screen_event,
            metadata_based_on_test_events_roi,
            metadata_based_on_test_events_full_screen,
        ),
    ],
)
def test_when_murko_called_with_full_screen_and_roi_event_then_metadata_updates_correctly(
    first_event,
    second_event,
    expected_md_after_first,
    expected_md_after_second,
    murko_callback: MurkoCallback,
):
    murko_callback.start(test_start_document)  # type: ignore
    murko_callback.event(first_event)
    murko_callback.event(test_smargon_event)

    murko_callback.redis_client.hset.assert_called_once_with(  # type: ignore
        "murko:12345:metadata",
        test_oav_uuid,
        json.dumps(expected_md_after_first),
    )
    murko_callback.redis_client.publish.assert_called_once_with(  # type: ignore
        "murko",
        json.dumps(expected_md_after_first),
    )

    murko_callback.event(second_event)
    murko_callback.event(test_smargon_event)

    assert murko_callback.redis_client.hset.call_args[0] == (  # type: ignore
        "murko:12345:metadata",
        test_oav_uuid,
        json.dumps(expected_md_after_second),
    )
    assert murko_callback.redis_client.publish.call_args[0] == (  # type: ignore
        "murko",
        json.dumps(expected_md_after_second),
    )


@pytest.mark.parametrize(
    "latest_omega, previous_omega, now, expected",
    [
        # Standard case
        (
            {"value": 10.0, "timestamp": 100.0},
            {"value": 5.0, "timestamp": 90.0},
            110.0,
            15.0,
        ),
        # Crossing zero from negative to positive
        (
            {"value": -5.0, "timestamp": 100.0},
            {"value": -15.0, "timestamp": 90.0},
            110.0,
            5.0,
        ),
        (
            {"value": 10.0, "timestamp": 100.0},
            {"value": -5.0, "timestamp": 90.0},
            110.0,
            25.0,
        ),
        # Crossing zero from positive to negative
        (
            {"value": 5.0, "timestamp": 100.0},
            {"value": 15.0, "timestamp": 90.0},
            110.0,
            -5.0,
        ),
        # Large time gaps between readings
        (
            {"value": 50.0, "timestamp": 500.0},
            {"value": 40.0, "timestamp": 400.0},
            600.0,
            60.0,
        ),
        # Very small time differences
        (
            {"value": 10.0, "timestamp": 100.0},
            {"value": 9.9, "timestamp": 99.99},
            100.01,
            10.1,
        ),
        # Identical omega values (should remain constant)
        (
            {"value": 30.0, "timestamp": 200.0},
            {"value": 30.0, "timestamp": 190.0},
            210.0,
            30.0,
        ),
    ],
)
def test_extrapolate_omega(latest_omega, previous_omega, now, expected):
    assert extrapolate_omega(latest_omega, previous_omega, now) == expected


@patch(
    "mx_bluesky.beamlines.i04.callbacks.murko_callback.MurkoCallback._check_redis_connection"
)
def test_if_redis_connection_fails_then_there_is_no_error(
    mock_check_redis_connection: MagicMock,
):
    mock_check_redis_connection.return_value = False
    callback = MurkoCallback("", "")
    doc = {}
    callback.start(doc)  # type: ignore
    callback.event(doc)  # type: ignore
    callback.stop(doc)  # type: ignore


def test_warning_is_logged_if_redis_connection_fails(caplog):
    callback = MurkoCallback("", "")
    doc = {}
    callback.start(doc)  # type: ignore
    log_message = caplog.records[-1]
    assert log_message.levelname == "WARNING"
    assert "Failed to connect to redis: " in log_message.message



from mx_bluesky.common.parameters.components import get_param_version
from mx_bluesky.hyperion.blueapi.parameters import (
    LoadCentreCollectParams,
    load_centre_collect_to_internal,
)
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect

from ....conftest import raw_params_from_file


def test_map_external_to_internal_parameters(tmp_path):
    raw_params = raw_params_from_file(
        "tests/test_data/parameter_json_files/external_load_centre_collect_params.json",
        tmp_path,
    )
    external_params = LoadCentreCollectParams(**raw_params)
    expected_internal = LoadCentreCollect(
        **(raw_params | {"parameter_model_version": get_param_version()})
    )
    actual_internal = load_centre_collect_to_internal(external_params)
    assert expected_internal == actual_internal



from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.aithre_lasershaping.goniometer import Goniometer
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection

from mx_bluesky.beamlines.aithre_lasershaping.parameters.constants import CONST
from mx_bluesky.common.experiment_plans.pin_tip_centring_plan import (
    PinTipCentringComposite,
    pin_tip_centre_plan,
)


def aithre_pin_tip_centre(
    oav: OAV = inject("OAV"),
    gonio: Goniometer = inject("gonio"),
    pin_tip_detection: PinTipDetection = inject("pin_tip_detection"),
    tip_offset_microns: float = 0,
    oav_config_file: str = CONST.OAV_CENTRING_FILE,
) -> MsgGenerator:
    """
    A plan that use pin_tip_centre_plan from common for aithre
    """

    composite = PinTipCentringComposite(oav, gonio, pin_tip_detection)

    yield from pin_tip_centre_plan(
        composite=composite,
        tip_offset_microns=tip_offset_microns,
        oav_config_file=oav_config_file,
    )



import pytest
from dodal.beamlines import i24
from dodal.devices.beamlines.i24.aperture import Aperture
from dodal.devices.beamlines.i24.beamstop import Beamstop
from dodal.devices.beamlines.i24.dcm import DCM
from dodal.devices.beamlines.i24.dual_backlight import DualBacklight
from dodal.devices.beamlines.i24.focus_mirrors import (
    FocusMirrorsMode,
    HFocusMode,
    VFocusMode,
)
from dodal.devices.beamlines.i24.pmac import PMAC
from dodal.devices.hutch_shutter import (
    HUTCH_SAFE_FOR_OPERATIONS,
    InterlockedHutchShutter,
    ShutterDemand,
    ShutterState,
)
from dodal.devices.motors import YZStage
from ophyd_async.core import callback_on_mock_put, set_mock_value


@pytest.fixture
def shutter() -> InterlockedHutchShutter:
    shutter = i24.shutter.build(connect_immediately=True, mock=True)
    set_mock_value(
        shutter.interlock.status,  # type: ignore
        HUTCH_SAFE_FOR_OPERATIONS,
    )

    def set_status(value: ShutterDemand, *args, **kwargs):
        value_sta = ShutterState.OPEN if value == "Open" else ShutterState.CLOSED
        set_mock_value(shutter.status, value_sta)

    callback_on_mock_put(shutter.control, set_status)
    return shutter


@pytest.fixture
def backlight() -> DualBacklight:
    return i24.backlight.build(connect_immediately=True, mock=True)


@pytest.fixture
def pmac() -> PMAC:
    return i24.pmac.build(connect_immediately=True, mock=True)


@pytest.fixture
def detector_stage() -> YZStage:
    return i24.detector_motion.build(connect_immediately=True, mock=True)


@pytest.fixture
def aperture() -> Aperture:
    return i24.aperture.build(connect_immediately=True, mock=True)


@pytest.fixture
def beamstop() -> Beamstop:
    return i24.beamstop.build(connect_immediately=True, mock=True)


@pytest.fixture
def dcm() -> DCM:
    return i24.dcm.build(connect_immediately=True, mock=True)


@pytest.fixture
def mirrors() -> FocusMirrorsMode:
    mirrors: FocusMirrorsMode = i24.focus_mirrors.build(
        connect_immediately=True, mock=True
    )
    set_mock_value(mirrors.horizontal, HFocusMode.FOCUS_10)
    set_mock_value(mirrors.vertical, VFocusMode.FOCUS_10)
    return mirrors


@pytest.fixture()
def use_i24_beamline(monkeypatch):
    monkeypatch.setenv("BEAMLINE", "i24")



from unittest.mock import AsyncMock, MagicMock, call, patch

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pytest
from bluesky import FailedStatus
from bluesky.callbacks import CallbackBase
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from ophyd_async.core import completed_status
from ophyd_async.fastcs.jungfrau import (
    AcquisitionType,
    GainMode,
    PedestalMode,
    create_jungfrau_internal_triggering_info,
)

from mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.do_darks import (
    do_non_pedestal_darks,
    do_pedestal_darks,
)


class CheckMonitor(CallbackBase):
    """Store the order and values of updates to specified signals

    Usage: Instantiate this callback with list of signals to track, and subscribe the run_engine to this
    callback. Run your plan using Bluesky's monitor_during decorator or wrapper, specifying the same signals
    in the monitor.
    """

    def __init__(self, signals_to_track: list[str]):
        self.signals_and_values = {signal: [] for signal in signals_to_track}

    def event(self, doc):
        key, value = next(iter(doc["data"].items()))
        # don't record a value changing to the same value
        if (
            not len(self.signals_and_values[key])
            or not self.signals_and_values[key][-1] == value
        ):
            self.signals_and_values[key].append(value)
        return doc


def fake_complete(_, group=None):
    yield from bps.null()
    return completed_status()


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils.log_on_percentage_complete",
    new=MagicMock,
)
@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils.bps.complete",
    new=MagicMock(side_effect=fake_complete),
)
async def test_full_do_pedestal_darks(
    jungfrau: CommissioningJungfrau,
    run_engine: RunEngine,
):
    # Test that plan succeeds in RunEngine and pedestal-specific signals are changed as expected
    test_path = "path"

    @bpp.run_decorator(
        md={
            "metadata": {"sample_id": "blah"},
            "activate_callbacks": ["SampleHandlingCallback"],
        }
    )
    def test_plan():
        yield from bps.monitor(jungfrau.drv.acquisition_type, name="AT")
        yield from bps.monitor(jungfrau.drv.pedestal_mode_state, name="PM")
        yield from bps.monitor(jungfrau.drv.gain_mode, name="GM")
        yield from do_pedestal_darks(0.001, 2, 2, test_path, jungfrau=jungfrau)

    jungfrau._controller.arm = AsyncMock()
    assert await jungfrau.drv.acquisition_type.get_value() == AcquisitionType.STANDARD
    await jungfrau.drv.gain_mode.set(GainMode.FIX_G2)
    await jungfrau.drv.pedestal_mode_state.set(PedestalMode.OFF)
    monitor_tracker = CheckMonitor(
        [
            "detector-drv-acquisition_type",
            "detector-drv-pedestal_mode_state",
            "detector-drv-gain_mode",
        ]
    )
    run_engine.subscribe(monitor_tracker)
    run_engine(test_plan())

    assert monitor_tracker.signals_and_values["detector-drv-acquisition_type"] == [
        AcquisitionType.STANDARD,
        AcquisitionType.PEDESTAL,
        AcquisitionType.STANDARD,
    ]
    assert monitor_tracker.signals_and_values["detector-drv-pedestal_mode_state"] == [
        PedestalMode.OFF,
        PedestalMode.ON,
        PedestalMode.OFF,
    ]

    # When using the real detector, the switching of gain mode is a bit more complicated,
    # see the docstring for the do_pedestal_darks plan.
    assert monitor_tracker.signals_and_values["detector-drv-gain_mode"] == [
        GainMode.FIX_G2,
        GainMode.DYNAMIC,
    ]


class FakeError(Exception): ...


async def test_pedestals_unstage_and_wait_on_exception(
    jungfrau: CommissioningJungfrau,
    run_engine: RunEngine,
):
    jungfrau.prepare = MagicMock(side_effect=FakeError)
    jungfrau.unstage = MagicMock(side_effect=lambda: completed_status())

    with pytest.raises(FakeError):
        run_engine(do_pedestal_darks(0.001, 2, 2, jungfrau=jungfrau))

    assert jungfrau.unstage.call_count == 1
    assert [c == call(jungfrau, wait=True) for c in jungfrau.unstage.call_args_list]


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils.log_on_percentage_complete",
    new=MagicMock(),
)
async def test_do_pedestals_waits_on_stage_before_prepare(
    jungfrau: CommissioningJungfrau, sim_run_engine: RunEngineSimulator
):
    msgs = sim_run_engine.simulate_plan(
        do_pedestal_darks(0.001, 2, 2, jungfrau=jungfrau)
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "stage" and msg.obj == jungfrau
    )
    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "wait")
    assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "prepare" and msg.obj == jungfrau
    )


def test_do_darks_stops_if_exception_after_stage(
    run_engine: RunEngine, jungfrau: CommissioningJungfrau
):
    mock_stop = AsyncMock()
    jungfrau.drv.acquisition_stop.trigger = mock_stop

    with pytest.raises(FailedStatus):
        run_engine(do_pedestal_darks(0, 2, 2, jungfrau=jungfrau))
    assert mock_stop.await_count == 2  # once when staging, once on exception
    assert [c == call(jungfrau, wait=True) for c in mock_stop.call_args_list]


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.do_darks.create_jungfrau_internal_triggering_info",
    new=MagicMock(side_effect=FakeError),
)
def test_do_non_pedestal_darks_unstages_jf_on_exception(
    run_engine: RunEngine, jungfrau: CommissioningJungfrau
):
    jungfrau.stage = MagicMock(side_effect=lambda: completed_status())
    jungfrau.unstage = MagicMock(side_effect=lambda: completed_status())
    with pytest.raises(FakeError):
        run_engine(do_non_pedestal_darks(GainMode.DYNAMIC, jungfrau=jungfrau))

    assert jungfrau.unstage.call_count == 1
    assert [c == call(jungfrau, wait=True) for c in jungfrau.unstage.call_args_list]


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.do_darks.fly_jungfrau"
)
def test_do_non_pedestal_darks_triggers_correct_plans(
    mock_fly_jf: MagicMock,
    run_engine: RunEngine,
    jungfrau: CommissioningJungfrau,
):
    gain_mode = GainMode.FORCE_SWITCH_G1
    parent_mock = MagicMock()
    jungfrau.unstage = MagicMock(side_effect=lambda: completed_status())
    jungfrau.stage = MagicMock(side_effect=lambda: completed_status())
    parent_mock.attach_mock(mock_fly_jf, "mock_fly_jf")
    parent_mock.attach_mock(jungfrau.unstage, "jungfrau_unstage")
    parent_mock.attach_mock(jungfrau.stage, "jungfrau_stage")
    expected_trigger_info = create_jungfrau_internal_triggering_info(1000, 0.001)
    run_engine(do_non_pedestal_darks(gain_mode=gain_mode, jungfrau=jungfrau))

    assert parent_mock.method_calls == [
        call.jungfrau_stage(),
        call.mock_fly_jf(
            jungfrau,
            expected_trigger_info,
            gain_mode,
            wait=True,
            log_on_percentage_prefix=f"Jungfrau {gain_mode} gain mode darks triggers received",
        ),
        call.jungfrau_unstage(),
    ]



from datetime import UTC, datetime, timedelta
from enum import StrEnum
from typing import Protocol
from urllib.parse import quote, urlencode


class Metadata(StrEnum):
    """Metadata fields that can be specified by the caller when raising an alert."""

    CONTAINER = "container"
    SAMPLE_ID = "sample_id"
    VISIT = "visit"


class ExtraMetadata(StrEnum):
    """Additional metadata fields that are automatically appended by
    the AlertService implementations."""

    ALERT_CONTENT = "alert_content"
    ALERT_SUMMARY = "alert_summary"
    BEAMLINE = "beamline"
    GRAYLOG_URL = "graylog_url"
    ISPYB_URL = "ispyb_url"
    PROPOSAL = "proposal"


class AlertService(Protocol):
    """
    Implemented by any backend that provides the ability to dispatch alerts to some
    service that is capable of disseminating them via any of a variety of media such
    as email, SMS, instant messaging, etc etc.
    """

    def raise_alert(self, summary: str, content: str, metadata: dict[Metadata, str]):
        """
        Raise an alert that will be forwarded to beamline support staff, which might
        for example be used as the basis for an incident in an incident reporting system.
        Args:
            summary: One line summary of the alert, that might for instance be used
                in an email subject line.
            content: Plain text content detailing the nature of the incident.
            metadata: A dict of strings that can be included as metadata in the alert for
                those backends that support it. The summary and content will be included
                by default.
        """
        pass


_alert_service: AlertService


def get_alerting_service() -> AlertService:
    """Get the alert service for this instance."""
    return _alert_service


def set_alerting_service(service: AlertService):
    """Set the alert service for this instance, call when the beamline is initialised."""
    global _alert_service
    _alert_service = service


def ispyb_url(sample_id: str):
    return f"https://ispyb.diamond.ac.uk/samples/sid/{quote(sample_id)}"


def graylog_url(stream_id: str):
    now = datetime.now(UTC)
    from_utc = now - timedelta(minutes=5)
    from_timestamp = from_utc.isoformat()
    # Add 1 second for graylog timing jitter
    to_utc = now + timedelta(seconds=1)
    to_timestamp = to_utc.isoformat()
    query_string = urlencode(
        {
            "streams": stream_id,
            "rangetype": "absolute",
            "from": from_timestamp,
            "to": to_timestamp,
        }
    )
    return "https://graylog.diamond.ac.uk/search?" + query_string



from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Sequence
from functools import partial

from mx_bluesky.common.utils.xrc_result import XRayCentreResult
from mx_bluesky.hyperion.blueapi.mixins import (
    MultiXtalSelection,
    TopNByMaxCountForEachSampleSelection,
    TopNByMaxCountSelection,
)


def top_n_by_max_count(
    unfiltered: Sequence[XRayCentreResult], n: int
) -> Sequence[XRayCentreResult]:
    sorted_hits = sorted(unfiltered, key=lambda result: result.max_count, reverse=True)
    return sorted_hits[:n]


def top_n_by_max_count_for_each_sample(
    unfiltered: Sequence[XRayCentreResult], n: int
) -> Sequence[XRayCentreResult]:
    xrc_results_by_sample_id: dict[int | None, list[XRayCentreResult]] = defaultdict(
        list[XRayCentreResult]
    )
    for result in unfiltered:
        xrc_results_by_sample_id[result.sample_id].append(result)
    return [
        result
        for results in xrc_results_by_sample_id.values()
        for result in sorted(results, key=lambda x: x.max_count, reverse=True)[:n]
    ]


def resolve_selection_fn(
    params: MultiXtalSelection,
) -> Callable[[Sequence[XRayCentreResult]], Sequence[XRayCentreResult]]:
    match params:
        case TopNByMaxCountSelection():
            return partial(top_n_by_max_count, n=params.n)
        case TopNByMaxCountForEachSampleSelection():
            return partial(top_n_by_max_count_for_each_sample, n=params.n)
    raise ValueError(f"Invalid selection function {params.name}")



from datetime import datetime
from unittest.mock import MagicMock, patch

import pytest
import requests
from bluesky import RunEngine

from mx_bluesky.beamlines.i24.serial.parameters.experiment_parameters import (
    ExtruderParameters,
    FixedTargetParameters,
)
from mx_bluesky.beamlines.i24.serial.write_nexus import call_nexgen, submit_to_server


@patch("mx_bluesky.beamlines.i24.serial.write_nexus.bps.sleep", MagicMock())
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.caget")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.cagetstring")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.read_text")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.exists")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.requests")
def test_call_nexgen_for_extruder(
    patch_request,
    fake_path,
    fake_read_text,
    fake_caget_str,
    fake_caget,
    dummy_params_ex: ExtruderParameters,
    run_engine: RunEngine,
):
    fake_caget_str.return_value = f"{dummy_params_ex.filename}_5001"
    fake_caget.return_value = 32
    fake_path.return_value = True
    fake_read_text.return_value = ""
    fake_start_time = datetime(2000, 1, 1)

    run_engine(call_nexgen(None, dummy_params_ex, 0.6, (1000, 1200), fake_start_time))
    patch_request.post.assert_called_once()

    nexgen_args = patch_request.post.call_args.kwargs["json"]
    assert nexgen_args["expt_type"] == "extruder"
    assert nexgen_args["start_time"] == fake_start_time.isoformat()


@patch("mx_bluesky.beamlines.i24.serial.write_nexus.bps.sleep", MagicMock())
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.caget")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.cagetstring")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.read_text")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.exists")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.requests")
def test_call_nexgen_for_fixed_target(
    patch_request,
    fake_path,
    fake_read_text,
    fake_caget_str,
    fake_caget,
    dummy_params_without_pp: FixedTargetParameters,
    run_engine: RunEngine,
):
    expected_filename = f"{dummy_params_without_pp.filename}_5002"
    fake_caget_str.return_value = expected_filename
    fake_caget.return_value = 32
    fake_path.return_value = True
    fake_read_text.return_value = ""
    fake_start_time = datetime(2000, 1, 1)
    run_engine(
        call_nexgen(None, dummy_params_without_pp, 0.6, (1000, 1200), fake_start_time)
    )
    patch_request.post.assert_called_once()

    nexgen_args = patch_request.post.call_args.kwargs["json"]
    assert nexgen_args["expt_type"] == "fixed-target"
    assert nexgen_args["filename"] == expected_filename
    assert nexgen_args["start_time"] == fake_start_time.isoformat()


@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.exists")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.read_text")
@patch(
    "mx_bluesky.beamlines.i24.serial.write_nexus.requests.post",
    side_effect=requests.HTTPError("No connection"),
)
def test_submit_to_nexgen_server_raises_http_error(
    fake_post,
    fake_read_text,
    fake_path,
):
    fake_path.return_value = True
    fake_read_text.return_value = ""

    with pytest.raises(requests.HTTPError, match="No connection"):
        submit_to_server(None)


@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.exists")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.read_text")
@patch(
    "mx_bluesky.beamlines.i24.serial.write_nexus.requests.post",
    side_effect=ValueError("Invalid payload"),
)
def test_submit_to_nexgen_server_raises_value_error(
    fake_post,
    fake_read_text,
    fake_path,
):
    fake_path.return_value = True
    fake_read_text.return_value = ""

    with pytest.raises(ValueError, match="Invalid payload"):
        submit_to_server(None)



from blueapi.core import BlueskyContext
from dodal.common.beamlines.beamline_utils import clear_devices
from dodal.common.beamlines.config_client import get_config_client
from dodal.utils import collect_factories, get_beamline_based_on_environment_variable


def setup_context(dev_mode: bool = False) -> BlueskyContext:
    context = BlueskyContext()
    setup_devices(context, dev_mode)
    return context


def clear_all_device_caches(context: BlueskyContext):
    context.unregister_all_devices()
    clear_devices()

    for f in collect_factories(get_beamline_based_on_environment_variable()).values():
        if hasattr(f, "cache_clear"):
            f.cache_clear()  # type: ignore

    get_config_client("i03").reset_cache()


def setup_devices(context: BlueskyContext, dev_mode: bool):
    _, exceptions = context.with_device_manager(
        get_beamline_based_on_environment_variable().devices,
        mock=dev_mode,
    )
    if exceptions:
        raise ExceptionGroup(
            f"Unable to connect to beamline devices {list(exceptions.keys())}",
            list(exceptions.values()),
        )



import os

from pydantic.dataclasses import dataclass

TEST_MODE = os.environ.get("AITHRE_TEST_MODE")


@dataclass(frozen=True)
class AithreConstants:
    BEAMLINE = "aithre"
    OAV_CENTRING_FILE = (
        "/dls/science/groups/i23/aithre/daq_configuration/json/OAVCentring_aithre.json"
    )
    LOG_FILE_NAME = "aithre.log"


CONST = AithreConstants()



from unittest.mock import MagicMock, Mock

import pytest
from blueapi.core import DataEvent
from bluesky_stomp.messaging import MessageContext, StompClient

from mx_bluesky.hyperion.external_interaction.callbacks.stomp.dispatcher import (
    StompDispatcher,
)

TEST_TASK_ID = "TEST_TASK_ID"

TEST_SUB_ID = "SUB_ID"


@pytest.fixture
def mock_client():
    stomp_client = Mock(spec=StompClient)
    stomp_client.subscribe.return_value = TEST_SUB_ID
    return stomp_client


def test_dispatcher_connect_and_disconnects_on_enter_and_exit(mock_client):
    dispatcher = StompDispatcher(mock_client)
    with dispatcher:
        mock_client.connect.assert_called_once()
        mock_client.subscribe.assert_called_once()
    mock_client.unsubscribe.assert_called_once_with(TEST_SUB_ID)
    mock_client.disconnect.assert_called_once()


def test_dispatcher_forwards_data_event(mock_client, test_event_data):
    dispatcher = StompDispatcher(mock_client)
    callback = MagicMock()

    with dispatcher:
        stomp_subs_callback = mock_client.subscribe.mock_calls[0].args[1]
        dispatcher.subscribe(callback)
        stomp_subs_callback(
            DataEvent(
                name="start",
                doc=test_event_data.test_grid_detect_and_gridscan_start_document,
                task_id=TEST_TASK_ID,
            ),
            Mock(spec=MessageContext),
        )
        callback.assert_called_once_with(
            "start", test_event_data.test_grid_detect_and_gridscan_start_document
        )



from functools import partial
from unittest.mock import MagicMock, patch

import pytest
from bluesky import plan_stubs as bps
from bluesky import preprocessors as bpp
from bluesky.preprocessors import run_decorator
from bluesky.run_engine import RunEngine
from dodal.devices.backlight import Backlight
from dodal.devices.robot import BartRobot
from ophyd_async.core import set_mock_value

from mx_bluesky.common.external_interaction.alerting import Metadata
from mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy import (
    robot_load_and_snapshots,
)
from mx_bluesky.hyperion.external_interaction.callbacks.alert_on_container_change import (
    AlertOnContainerChange,
)
from mx_bluesky.hyperion.parameters.constants import CONST

TEST_SAMPLE_ID = 10
TEST_VISIT = "cm1234-67"
TEST_CONTAINER = 8


def wrap_plan(container: int, plan):
    @bpp.set_run_key_decorator("inner_subplan")
    @run_decorator(md={"metadata": {"some_other_metadata": 1234}})
    def inner_subplan():
        yield from plan()

    yield from bpp.run_wrapper(
        inner_subplan(),
        md={
            "metadata": {
                "container": container,
                "sample_id": TEST_SAMPLE_ID,
                "visit": TEST_VISIT,
            },
            "activate_callbacks": ["AlertOnContainerChange"],
        },
    )


def dummy_plan_with_container(robot: BartRobot, container: int):
    def my_plan():
        yield from bps.create(name=CONST.DESCRIPTORS.ROBOT_PRE_LOAD)
        yield from bps.read(robot)
        yield from bps.save()

    yield from wrap_plan(container, my_plan)


@patch.dict("os.environ", {"BEAMLINE": "i03"})
def test_when_data_collected_on_the_same_container_then_does_not_alert_multiple_times(
    run_engine: RunEngine, mock_alert_service: MagicMock, robot: BartRobot
):
    run_engine.subscribe(AlertOnContainerChange())

    set_mock_value(robot.current_puck, 5)

    run_engine(dummy_plan_with_container(robot, 5))
    run_engine(dummy_plan_with_container(robot, 5))
    run_engine(dummy_plan_with_container(robot, 5))

    mock_alert_service.raise_alert.assert_not_called()


@patch.dict("os.environ", {"BEAMLINE": "i03"})
def test_when_data_collected_on_new_container_then_alerts(
    run_engine: RunEngine,
    mock_alert_service: MagicMock,
    robot: BartRobot,
):
    run_engine.subscribe(AlertOnContainerChange())
    set_mock_value(robot.current_puck, 5)

    run_engine(dummy_plan_with_container(robot, 10))

    mock_alert_service.raise_alert.assert_called_once_with(
        "UDC moved on to puck 10 on i03",
        "Hyperion finished container 5 and moved on to 10",
        {
            Metadata.SAMPLE_ID: "10",
            Metadata.VISIT: "cm1234-67",
            Metadata.CONTAINER: "10",
        },
    )


@patch.dict("os.environ", {"BEAMLINE": "i03"})
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.do_robot_load"
)
def test_robot_load_and_snapshots_triggers_alert_and_resets_state_between_plans(
    patched_robot_load: MagicMock,
    run_engine: RunEngine,
    mock_alert_service: MagicMock,
    robot: BartRobot,
    backlight: Backlight,
):
    run_engine.subscribe(AlertOnContainerChange())

    mock_composite = MagicMock()
    mock_composite.robot = robot
    mock_composite.backlight = backlight

    patched_robot_load.side_effect = NotImplementedError()

    old_container = 5
    for container in range(1, 3):
        set_mock_value(robot.current_puck, old_container)
        with pytest.raises(NotImplementedError):
            run_engine(
                wrap_plan(
                    container,
                    partial(
                        robot_load_and_snapshots,
                        mock_composite,
                        MagicMock(),
                        MagicMock(),
                        1,
                        None,
                    ),
                )
            )

        mock_alert_service.raise_alert.assert_called_once_with(
            f"UDC moved on to puck {container} on i03",
            f"Hyperion finished container {old_container} and moved on to {container}",
            {
                Metadata.SAMPLE_ID: "10",
                Metadata.VISIT: "cm1234-67",
                Metadata.CONTAINER: str(container),
            },
        )
        old_container = container
        mock_alert_service.reset_mock()



from contextlib import nullcontext as does_not_raise
from pathlib import Path

import pytest
from pydantic import BaseModel, ValidationError

from mx_bluesky.common.parameters.components import (
    MxBlueskyParameters,
    WithSnapshot,
    get_param_version,
)


@pytest.mark.parametrize(
    "model, expectation",
    [
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [],
                "use_grid_snapshots": False,
            },
            does_not_raise(),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [],
                "use_grid_snapshots": True,
            },
            does_not_raise(),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [10, 20, 30, 40],
                "use_grid_snapshots": False,
            },
            does_not_raise(),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [0, 270],
                "use_grid_snapshots": True,
            },
            pytest.raises(ValidationError),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [0],
                "use_grid_snapshots": True,
            },
            pytest.raises(ValidationError),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [10, 80],
                "use_grid_snapshots": True,
            },
            pytest.raises(ValidationError),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "use_grid_snapshots": True,
            },
            does_not_raise(),
        ],
    ],
)
def test_validate_with_snapshot_omegas_grid_snapshots(model, expectation):
    with expectation:
        WithSnapshot.model_validate(model)


class PlainModel(MxBlueskyParameters):
    outer_path: Path


class InnerModel(BaseModel):
    inner_path: Path


class NestedModel(MxBlueskyParameters):
    outer_path: Path
    inner_model: InnerModel


def test_mx_bluesky_parameters_round_trip_python():
    params = MxBlueskyParameters(parameter_model_version=get_param_version())
    serialized_model = params.model_dump()
    new_params = MxBlueskyParameters.model_validate(serialized_model)
    assert new_params == params


def test_mx_bluesky_parameters_path_round_trip_python():
    params = PlainModel(
        parameter_model_version=get_param_version(), outer_path=Path("/tmp")
    )
    serialized_model = params.model_dump()
    new_params = PlainModel.model_validate(serialized_model)
    assert new_params == params


def test_mx_bluesky_parameters_nested_path_round_trip_python():
    params = NestedModel(
        parameter_model_version=get_param_version(),
        outer_path=Path("/tmp"),
        inner_model=InnerModel(inner_path=Path("/tmp")),
    )
    serialized_model = params.model_dump()
    new_params = NestedModel.model_validate(serialized_model)
    assert new_params == params


def test_mx_bluesky_parameters_round_trip_json():
    params = MxBlueskyParameters(parameter_model_version=get_param_version())
    serialized_model = params.model_dump_json()
    new_params = MxBlueskyParameters.model_validate_json(serialized_model)
    assert new_params == params



from __future__ import annotations

from collections.abc import Callable, Sequence
from enum import StrEnum
from math import isclose
from time import time
from typing import TYPE_CHECKING, Any, TypeVar

from bluesky import preprocessors as bpp
from bluesky.utils import MsgGenerator, make_decorator
from dodal.devices.zocalo import ZocaloStartInfo

from mx_bluesky.common.external_interaction.callbacks.common.ispyb_callback_base import (
    BaseISPyBCallback,
    D,
)
from mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping import (
    populate_data_collection_group,
    populate_remaining_data_collection_info,
)
from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloInfoGenerator,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_mapping import (
    construct_comment_for_gridscan,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    DataCollectionGroupInfo,
    DataCollectionInfo,
    DataCollectionPositionInfo,
    Orientation,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.parameters.components import DiffractionExperimentWithSample
from mx_bluesky.common.parameters.constants import DocDescriptorNames, PlanNameConstants
from mx_bluesky.common.parameters.gridscan import GridCommon
from mx_bluesky.common.utils.exceptions import (
    ISPyBDepositionNotMadeError,
    SampleError,
)
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER, set_dcgid_tag
from mx_bluesky.common.utils.utils import number_of_frames_from_scan_spec

OMEGA_TOLERANCE = 1


class GridscanPlane(StrEnum):
    OMEGA_XY = "0"
    OMEGA_XZ = "90"


if TYPE_CHECKING:
    from event_model import Event, RunStart, RunStop

T = TypeVar("T", bound="GridCommon")
ASSERT_START_BEFORE_EVENT_DOC_MESSAGE = f"No data collection group info - event document has been emitted before a {PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN} start document"


def ispyb_activation_wrapper(plan_generator: MsgGenerator, parameters):
    return bpp.set_run_key_wrapper(
        bpp.run_wrapper(
            plan_generator,
            md={
                "activate_callbacks": ["GridscanISPyBCallback"],
                "subplan_name": PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN,
                "mx_bluesky_parameters": parameters.model_dump_json(),
            },
        ),
        PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN,
    )


ispyb_activation_decorator = make_decorator(ispyb_activation_wrapper)


class GridscanISPyBCallback(BaseISPyBCallback):
    """Callback class to handle the deposition of experiment parameters into the ISPyB
    database. Listens for 'event' and 'descriptor' documents. Creates the ISpyB entry on
    receiving an 'event' document for the 'ispyb_reading_hardware' event, and updates the
    deposition on receiving its final 'stop' document.

    To use, subscribe the Bluesky RunEngine to an instance of this class.
    E.g.:
        ispyb_handler_callback = FGSISPyBCallback(parameters)
        run_engine.subscribe(ispyb_handler_callback)
    Or decorate a plan using bluesky.preprocessors.subs_decorator.

    See: https://blueskyproject.io/bluesky/callbacks.html#ways-to-invoke-callbacks
    """

    def __init__(
        self,
        param_type: type[T],
        *,
        emit: Callable[..., Any] | None = None,
    ) -> None:
        super().__init__(emit=emit)
        self.ispyb: StoreInIspyb
        self.param_type = param_type
        self._start_of_fgs_uid: str | None = None
        self._processing_start_time: float | None = None
        self._grid_plane_to_id_map: dict[GridscanPlane, int] = {}
        self._grid_plane_to_width_map: dict[GridscanPlane, int] = {}
        self.data_collection_group_info: DataCollectionGroupInfo | None

    def activity_gated_start(self, doc: RunStart):
        if doc.get("subplan_name") == PlanNameConstants.DO_FGS:
            self._start_of_fgs_uid = doc.get("uid")

        if doc.get("subplan_name") == PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN:
            self.uid_to_finalize_on = doc.get("uid")
            ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                "ISPyB callback received start document with experiment parameters and "
                f"uid: {self.uid_to_finalize_on}"
            )
            mx_bluesky_parameters = doc.get("mx_bluesky_parameters")
            assert isinstance(mx_bluesky_parameters, str)
            self.params = self.param_type.model_validate_json(mx_bluesky_parameters)
            assert isinstance(self.params, DiffractionExperimentWithSample)
            self.ispyb = StoreInIspyb(self.ispyb_config)
            self.data_collection_group_info = populate_data_collection_group(
                self.params
            )

            scan_data_infos = [
                ScanDataInfo(
                    data_collection_info=populate_remaining_data_collection_info(
                        "MX-Bluesky: Xray centring 1 -",
                        None,
                        DataCollectionInfo(),
                        self.params,
                    ),
                ),
                ScanDataInfo(
                    data_collection_info=populate_remaining_data_collection_info(
                        "MX-Bluesky: Xray centring 2 -",
                        None,
                        DataCollectionInfo(),
                        self.params,
                    )
                ),
            ]

            self.ispyb_ids = self.ispyb.begin_deposition(
                self.data_collection_group_info, scan_data_infos
            )
            set_dcgid_tag(self.ispyb_ids.data_collection_group_id)
        return super().activity_gated_start(doc)

    def activity_gated_event(self, doc: Event):
        assert self.data_collection_group_info, ASSERT_START_BEFORE_EVENT_DOC_MESSAGE

        doc = super().activity_gated_event(doc)

        descriptor_name = self.descriptors[doc["descriptor"]].get("name")
        if descriptor_name == DocDescriptorNames.OAV_GRID_SNAPSHOT_TRIGGERED:
            scan_data_infos = self._handle_oav_grid_snapshot_triggered(doc)
            self.ispyb_ids = self.ispyb.update_deposition(
                self.ispyb_ids, scan_data_infos
            )
        self.ispyb.update_data_collection_group_table(
            self.data_collection_group_info, self.ispyb_ids.data_collection_group_id
        )

        return doc

    def _add_processing_time_to_comment(self, processing_start_time: float):
        assert self.data_collection_group_info, ASSERT_START_BEFORE_EVENT_DOC_MESSAGE
        proc_time = time() - processing_start_time
        crystal_summary = f"Zocalo processing took {proc_time:.2f} s."

        self.data_collection_group_info.comments = (
            self.data_collection_group_info.comments or ""
        ) + crystal_summary

        self.ispyb.append_to_comment(
            self.ispyb_ids.data_collection_ids[0], crystal_summary
        )

    def _handle_oav_grid_snapshot_triggered(self, doc) -> Sequence[ScanDataInfo]:
        assert self.ispyb_ids.data_collection_ids, "No current data collection"
        assert self.params, "ISPyB handler didn't receive parameters!"
        assert self.data_collection_group_info, "No data collection group"
        data = doc["data"]
        omega = doc["data"]["gonio-omega"]
        grid_plane = _smargon_omega_to_xyxz_plane(omega)
        ISPYB_ZOCALO_CALLBACK_LOGGER.info(
            f"Generating dc info for gridplane {grid_plane}, omega {omega}"
        )
        data_collection_number = self.data_collection_number_from_gridplane(grid_plane)
        file_template = (
            f"{self.params.detector_params.prefix}_{data_collection_number}_master.h5"
        )
        data_collection_info = DataCollectionInfo(
            xtal_snapshot1=data.get("oav-grid_snapshot-last_path_full_overlay"),
            xtal_snapshot2=data.get("oav-grid_snapshot-last_path_outer"),
            xtal_snapshot3=data.get("oav-grid_snapshot-last_saved_path"),
            n_images=(
                data["oav-grid_snapshot-num_boxes_x"]
                * data["oav-grid_snapshot-num_boxes_y"]
            ),
            data_collection_number=data_collection_number,
            file_template=file_template,
        )
        microns_per_pixel_x = data["oav-microns_per_pixel_x"]
        microns_per_pixel_y = data["oav-microns_per_pixel_y"]
        data_collection_grid_info = DataCollectionGridInfo(
            dx_in_mm=data["oav-grid_snapshot-box_width"] * microns_per_pixel_x / 1000,
            dy_in_mm=data["oav-grid_snapshot-box_width"] * microns_per_pixel_y / 1000,
            steps_x=data["oav-grid_snapshot-num_boxes_x"],
            steps_y=data["oav-grid_snapshot-num_boxes_y"],
            microns_per_pixel_x=microns_per_pixel_x,
            microns_per_pixel_y=microns_per_pixel_y,
            snapshot_offset_x_pixel=int(data["oav-grid_snapshot-top_left_x"]),
            snapshot_offset_y_pixel=int(data["oav-grid_snapshot-top_left_y"]),
            orientation=Orientation.HORIZONTAL,
            snaked=True,
        )
        data_collection_info.comments = construct_comment_for_gridscan(
            data_collection_grid_info
        )

        # Snapshots may be triggered in a different order to gridscans, so save
        # the mapping to the data collection id in order to trigger Zocalo correctly.
        data_collection_id = self.ispyb_ids.data_collection_ids[
            0 if grid_plane == GridscanPlane.OMEGA_XY else 1
        ]
        self._grid_plane_to_id_map[grid_plane] = data_collection_id
        self._grid_plane_to_width_map[grid_plane] = data_collection_grid_info.steps_y

        y_steps = self._grid_plane_to_width_map.get(GridscanPlane.OMEGA_XY, "_")
        z_steps = self._grid_plane_to_width_map.get(GridscanPlane.OMEGA_XZ, "_")
        self.data_collection_group_info.comments = (
            f"Diffraction grid scan of {data_collection_grid_info.steps_x} by "
            f"{y_steps} by {z_steps}."
        )

        self._populate_axis_info(data_collection_info, doc["data"])

        scan_data_info = ScanDataInfo(
            data_collection_info=data_collection_info,
            data_collection_id=data_collection_id,
            data_collection_grid_info=data_collection_grid_info,
        )
        ISPYB_ZOCALO_CALLBACK_LOGGER.info(
            "Updating ispyb data collection after oav snapshot."
        )

        self._oav_snapshot_event_idx += 1
        return [scan_data_info]

    def _populate_axis_info(self, data_collection_info: DataCollectionInfo, doc: dict):
        if (omega_start := doc.get("gonio-omega")) is not None:
            omega_in_gda_space = -omega_start
            data_collection_info.omega_start = omega_in_gda_space
            data_collection_info.axis_start = omega_in_gda_space
            data_collection_info.axis_end = omega_in_gda_space
            data_collection_info.axis_range = 0
        if (chi_start := doc.get("gonio-chi")) is not None:
            data_collection_info.chi_start = chi_start

    def populate_info_for_update(
        self,
        event_sourced_data_collection_info: DataCollectionInfo,
        event_sourced_position_info: DataCollectionPositionInfo | None,
        params: DiffractionExperimentWithSample,
    ) -> Sequence[ScanDataInfo]:
        assert self.ispyb_ids.data_collection_ids, (
            "Expect at least one valid data collection to record scan data"
        )
        xy_scan_data_info = ScanDataInfo(
            data_collection_info=event_sourced_data_collection_info,
            data_collection_id=self.ispyb_ids.data_collection_ids[0],
        )
        scan_data_infos = [xy_scan_data_info]

        data_collection_id = (
            self.ispyb_ids.data_collection_ids[1]
            if len(self.ispyb_ids.data_collection_ids) > 1
            else None
        )
        xz_scan_data_info = ScanDataInfo(
            data_collection_info=event_sourced_data_collection_info,
            data_collection_id=data_collection_id,
        )
        scan_data_infos.append(xz_scan_data_info)
        return scan_data_infos

    def activity_gated_stop(self, doc: RunStop) -> RunStop:
        assert self.data_collection_group_info, (
            f"No data collection group info - stop document has been emitted before a {PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN} start document"
        )
        if doc.get("run_start") == self._start_of_fgs_uid:
            self._processing_start_time = time()
        if doc.get("run_start") == self.uid_to_finalize_on:
            ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                "ISPyB callback received stop document corresponding to start document "
                f"with uid: {self.uid_to_finalize_on}."
            )
            if self.ispyb_ids == IspybIds():
                raise ISPyBDepositionNotMadeError(
                    "ispyb was not initialised at run start"
                )
            exception_type, message = SampleError.type_and_message_from_reason(
                doc.get("reason", "")
            )
            if exception_type:
                doc["reason"] = message
                self.data_collection_group_info.comments = message
            elif self._processing_start_time:
                self._add_processing_time_to_comment(self._processing_start_time)
            self.ispyb.update_data_collection_group_table(
                self.data_collection_group_info,
                self.ispyb_ids.data_collection_group_id,
            )
            self.data_collection_group_info = None
            self._grid_plane_to_id_map.clear()
            self._grid_plane_to_width_map.clear()
            return super().activity_gated_stop(doc)
        return self.tag_doc(doc)

    def tag_doc(self, doc: D) -> D:
        doc = super().tag_doc(doc)
        assert isinstance(doc, dict)
        if self._grid_plane_to_id_map:
            doc["grid_plane_to_id_map"] = self._grid_plane_to_id_map
        return doc  # type: ignore

    def data_collection_number_from_gridplane(self, plane) -> int:
        assert self.params
        base_number = self.params.detector_params.run_number
        return base_number if plane == GridscanPlane.OMEGA_XY else base_number + 1


def generate_start_info_from_omega_map() -> ZocaloInfoGenerator:
    """
    Generate the zocalo trigger info from bluesky runs where the frame number is
    computed using metadata added to the document by the ISPyB callback and the
    run start which together can be used to determine the correct frame numbering.
    """
    doc = yield []
    omega_to_scan_spec = doc["omega_to_scan_spec"]
    start_frame = 0
    infos = []
    for i, omega in enumerate([GridscanPlane.OMEGA_XY, GridscanPlane.OMEGA_XZ]):
        frames = number_of_frames_from_scan_spec(omega_to_scan_spec[omega])
        infos.append(
            ZocaloStartInfo(
                doc["grid_plane_to_id_map"][omega], None, start_frame, frames, i
            )
        )
        start_frame += frames
    yield infos


def _smargon_omega_to_xyxz_plane(smargon_omega: float) -> GridscanPlane:
    modulo_180 = abs(smargon_omega) % 180
    is_xy = isclose(modulo_180, 0, abs_tol=OMEGA_TOLERANCE)
    assert is_xy or isclose(modulo_180, 90, abs_tol=OMEGA_TOLERANCE), (
        f"Smargon snapshot omega not in tolerance of compass point {smargon_omega}"
    )
    return GridscanPlane.OMEGA_XY if is_xy else GridscanPlane.OMEGA_XZ



import signal
from pathlib import Path
from sys import argv

from blueapi.config import ApplicationConfig, ConfigLoader
from blueapi.core import BlueskyContext

from mx_bluesky.common.external_interaction import alerting
from mx_bluesky.common.external_interaction.alerting.log_based_service import (
    LoggingAlertService,
)
from mx_bluesky.common.utils.log import (
    LOGGER,
    do_default_logging_setup,
)
from mx_bluesky.hyperion.baton_handler import run_forever
from mx_bluesky.hyperion.in_process_runner import InProcessRunner
from mx_bluesky.hyperion.parameters.cli import (
    HyperionArgs,
    HyperionMode,
    parse_cli_args,
)
from mx_bluesky.hyperion.parameters.constants import CONST, HyperionConstants
from mx_bluesky.hyperion.plan_runner import PlanRunner
from mx_bluesky.hyperion.plan_runner_api import create_server_for_udc
from mx_bluesky.hyperion.supervisor import SupervisorRunner
from mx_bluesky.hyperion.utils.context import setup_context


def initialise_globals(args: HyperionArgs):
    """Do all early main low-level application initialisation."""
    do_default_logging_setup(
        CONST.SUPERVISOR_LOG_FILE_NAME
        if args.mode == HyperionMode.SUPERVISOR
        else CONST.LOG_FILE_NAME,
        CONST.GRAYLOG_PORT,
        dev_mode=args.dev_mode,
    )
    LOGGER.info(f"Hyperion launched with args:{argv}")
    alerting.set_alerting_service(LoggingAlertService(CONST.GRAYLOG_STREAM_ID))


def main():
    """Main application entry point."""
    args = parse_cli_args()
    initialise_globals(args)
    server_port = HyperionConstants.HYPERION_PORT
    match args.mode:
        case HyperionMode.UDC:
            context = setup_context(dev_mode=args.dev_mode)
            plan_runner = InProcessRunner(context, args.dev_mode)
        case HyperionMode.SUPERVISOR:
            if not args.client_config:
                raise RuntimeError(
                    "BlueAPI client configuration file must be specified in supervisor mode."
                )
            if not args.supervisor_config:
                raise RuntimeError(
                    "BlueAPI supervisor configuration file must be specified in supervisor mode."
                )

            client_config = _load_config_from_yaml(Path(args.client_config))
            supervisor_config = _load_config_from_yaml(Path(args.supervisor_config))
            context = BlueskyContext(configuration=supervisor_config)
            plan_runner = SupervisorRunner(context, client_config, args.dev_mode)
            server_port = HyperionConstants.SUPERVISOR_PORT
    create_server_for_udc(plan_runner, server_port)
    _register_sigterm_handler(plan_runner)
    run_forever(plan_runner)


def _load_config_from_yaml(config_path: Path):
    loader = ConfigLoader(ApplicationConfig)
    loader.use_values_from_yaml(config_path)
    return loader.load()


def _register_sigterm_handler(runner: PlanRunner):
    def shutdown_on_sigterm(sig_num, frame):
        LOGGER.info("Received SIGTERM, shutting down...")
        runner.shutdown()

    signal.signal(signal.SIGTERM, shutdown_on_sigterm)


if __name__ == "__main__":
    main()



from functools import partial
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from bluesky.simulators import RunEngineSimulator
from bluesky.utils import Msg
from dodal.devices.beamsize.beamsize import BeamsizeBase
from dodal.devices.synchrotron import SynchrotronMode
from dodal.devices.zocalo import ZocaloResults
from ophyd_async.core import AsyncStatus, completed_status, set_mock_value

from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.utils.xrc_result import XRayCentreResult
from mx_bluesky.hyperion.experiment_plans.hyperion_flyscan_xray_centre_plan import (
    construct_hyperion_specific_features,
)
from mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy import (
    RobotLoadAndEnergyChangeComposite,
)
from mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan import (
    RobotLoadThenCentreComposite,
)
from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import (
    create_gridscan_callbacks,
)
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

FLYSCAN_RESULT_HIGH = XRayCentreResult(
    centre_of_mass_mm=np.array([0.1, 0.2, 0.3]),
    bounding_box_mm=(np.array([0.09, 0.19, 0.29]), np.array([0.11, 0.21, 0.31])),
    max_count=30,
    total_count=100,
    sample_id=2,
)
FLYSCAN_RESULT_MED = XRayCentreResult(
    centre_of_mass_mm=np.array([0.4, 0.5, 0.6]),
    bounding_box_mm=(np.array([0.09, 0.19, 0.29]), np.array([0.11, 0.21, 0.31])),
    max_count=20,
    total_count=120,
    sample_id=1,
)
FLYSCAN_RESULT_LOW = XRayCentreResult(
    centre_of_mass_mm=np.array([0.7, 0.8, 0.9]),
    bounding_box_mm=(np.array([0.09, 0.19, 0.29]), np.array([0.11, 0.21, 0.31])),
    max_count=10,
    total_count=140,
    sample_id=1,
)
FLYSCAN_RESULT_HIGH_NO_SAMPLE_ID = XRayCentreResult(
    centre_of_mass_mm=np.array([0.1, 0.2, 0.3]),
    bounding_box_mm=(np.array([0.09, 0.19, 0.29]), np.array([0.11, 0.21, 0.31])),
    max_count=30,
    total_count=100,
    sample_id=None,
)


@pytest.fixture(autouse=True)
def always_use_i03_beamline(use_beamline_i03): ...


@pytest.fixture
def sim_run_engine_for_rotation(sim_run_engine):
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"values": {"value": SynchrotronMode.USER}},
        "synchrotron-synchrotron_mode",
    )
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"values": {"value": -1}},
        "synchrotron-top_up_start_countdown",
    )
    sim_run_engine.add_handler(
        "read", lambda msg: {"values": {"value": -1}}, "gonio_omega"
    )
    return sim_run_engine


def mock_zocalo_trigger(zocalo: ZocaloResults, result):
    @AsyncStatus.wrap
    async def mock_complete(results):
        await zocalo._put_results(results, {"dcid": 0, "dcgid": 0})

    zocalo.trigger = MagicMock(side_effect=partial(mock_complete, result))


def modified_store_grid_scan_mock(*args, dcids=(0, 0), dcgid=0, **kwargs):
    mock = MagicMock(spec=StoreInIspyb)
    mock.begin_deposition.return_value = IspybIds(
        data_collection_ids=dcids, data_collection_group_id=dcgid
    )
    mock.update_deposition.return_value = IspybIds(
        data_collection_ids=dcids, data_collection_group_id=dcgid, grid_ids=(0, 0)
    )
    return mock


@pytest.fixture
def mock_subscriptions():
    with (
        patch(
            "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
            autospec=True,
        ),
        patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb.append_to_comment"
        ),
        patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb.begin_deposition",
            new=MagicMock(
                return_value=IspybIds(
                    data_collection_ids=(0, 0), data_collection_group_id=0
                )
            ),
        ),
        patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb.update_deposition",
            new=MagicMock(
                return_value=IspybIds(
                    data_collection_ids=(0, 0),
                    data_collection_group_id=0,
                    grid_ids=(0, 0),
                )
            ),
        ),
    ):
        nexus_callback, ispyb_callback = create_gridscan_callbacks()
        ispyb_callback.ispyb = MagicMock(spec=StoreInIspyb)

    return (nexus_callback, ispyb_callback)


def fake_read(obj, initial_positions, _):
    initial_positions[obj] = 0
    yield Msg("null", obj)


@pytest.fixture
def robot_load_composite(
    smargon,
    dcm,
    robot,
    aperture_scatterguard,
    oav,
    webcam,
    thawer,
    lower_gonio,
    eiger,
    xbpm_feedback,
    attenuator,
    beamstop_phase1,
    fast_grid_scan,
    undulator,
    undulator_dcm,
    s4_slit_gaps,
    vfm,
    mirror_voltages,
    backlight,
    detector_motion,
    flux,
    pin_tip_detection_with_found_pin,
    zocalo,
    synchrotron,
    sample_shutter,
    zebra,
    panda,
    panda_fast_grid_scan,
    beamsize: BeamsizeBase,
) -> RobotLoadThenCentreComposite:
    set_mock_value(dcm.energy_in_keV.user_readback, 11.105)
    smargon.stub_offsets.set = MagicMock(side_effect=lambda _: completed_status())
    aperture_scatterguard.set = MagicMock(side_effect=lambda _: completed_status())
    set_mock_value(smargon.omega.max_velocity, 131)
    return RobotLoadThenCentreComposite(
        xbpm_feedback=xbpm_feedback,
        attenuator=attenuator,
        aperture_scatterguard=aperture_scatterguard,
        backlight=backlight,
        beamsize=beamsize,
        beamstop=beamstop_phase1,
        detector_motion=detector_motion,
        eiger=eiger,
        zebra_fast_grid_scan=fast_grid_scan,
        flux=flux,
        oav=oav,
        pin_tip_detection=pin_tip_detection_with_found_pin,
        gonio=smargon,
        synchrotron=synchrotron,
        s4_slit_gaps=s4_slit_gaps,
        undulator=undulator,
        zebra=zebra,
        zocalo=zocalo,
        panda=panda,
        panda_fast_grid_scan=panda_fast_grid_scan,
        thawer=thawer,
        sample_shutter=sample_shutter,
        vfm=vfm,
        mirror_voltages=mirror_voltages,
        dcm=dcm,
        undulator_dcm=undulator_dcm,
        robot=robot,
        webcam=webcam,
        lower_gonio=lower_gonio,
    )


@pytest.fixture
def robot_load_and_energy_change_composite(
    smargon,
    dcm,
    robot,
    aperture_scatterguard,
    oav,
    webcam,
    thawer,
    lower_gonio,
    vfm,
    mirror_voltages,
    undulator_dcm,
    xbpm_feedback,
    attenuator,
    backlight,
) -> RobotLoadAndEnergyChangeComposite:
    composite = RobotLoadAndEnergyChangeComposite(
        vfm,
        mirror_voltages,
        dcm,
        undulator_dcm,
        xbpm_feedback,
        attenuator,
        robot,
        webcam,
        lower_gonio,
        thawer,
        oav,
        smargon,
        aperture_scatterguard,
        backlight,
    )
    composite.gonio.stub_offsets.set = MagicMock(
        side_effect=lambda _: completed_status()
    )
    composite.aperture_scatterguard.set = MagicMock(
        side_effect=lambda _: completed_status()
    )
    set_mock_value(composite.dcm.energy_in_keV.user_readback, 11.105)

    return composite


def sim_fire_event_on_open_run(sim_run_engine: RunEngineSimulator, run_name: str):
    def fire_event(msg: Msg):
        try:
            sim_run_engine.fire_callback("start", msg.kwargs)
        except Exception as e:
            print(f"Exception is {e}")

    def msg_maches_run(msg: Msg):
        return msg.run == run_name

    sim_run_engine.add_handler("open_run", fire_event, msg_maches_run)


@pytest.fixture
def grid_detection_callback_with_detected_grid():
    with patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
        autospec=True,
    ) as callback:
        callback.return_value.get_grid_parameters.return_value = {
            "transmission_frac": 1.0,
            "exposure_time_s": 0,
            "x_start_um": 0,
            "y_start_um": 0,
            "y2_start_um": 0,
            "z_start_um": 0,
            "z2_start_um": 0,
            "x_steps": 10,
            "y_steps": 10,
            "z_steps": 10,
            "x_step_size_um": 0.1,
            "y_step_size_um": 0.1,
            "z_step_size_um": 0.1,
        }
        yield callback


@pytest.fixture
def beamline_specific(
    hyperion_flyscan_xrc_composite: HyperionFlyScanXRayCentreComposite,
    hyperion_fgs_params: HyperionSpecifiedThreeDGridScan,
) -> BeamlineSpecificFGSFeatures:
    return construct_hyperion_specific_features(
        hyperion_flyscan_xrc_composite, hyperion_fgs_params
    )



from dodal.devices.beamlines.i02_1.fast_grid_scan import ZebraGridScanParamsTwoD
from scanspec.specs import Product

from mx_bluesky.common.parameters.components import SplitScan, WithOptionalEnergyChange
from mx_bluesky.common.parameters.gridscan import SpecifiedGrid


class SpecifiedTwoDGridScan(
    SpecifiedGrid[ZebraGridScanParamsTwoD],
    SplitScan,
    WithOptionalEnergyChange,
):
    """Parameters representing a so-called 2D grid scan, which consists of doing a
    gridscan in X and Y."""

    @property
    def scan_spec(self) -> Product[str]:
        """A fully specified ScanSpec object representing the grid, with x, y, z and
        omega positions."""
        return self.grid_1_spec

    @property
    def fast_gridscan_params(self) -> ZebraGridScanParamsTwoD:
        return ZebraGridScanParamsTwoD(
            x_steps=self.x_steps,
            y_steps=self.y_steps,
            x_step_size_mm=self.x_step_size_um / 1000,
            y_step_size_mm=self.y_step_size_um / 1000,
            x_start_mm=self.x_start_um / 1000,
            y1_start_mm=self.y_start_um / 1000,
            z1_start_mm=self.z_start_um / 1000,
            set_stub_offsets=self._set_stub_offsets,
            transmission_fraction=0.5,
            dwell_time_ms=self.exposure_time_s,
        )
